
import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

import nipype.interfaces.fsl as fsl
import nipype.interfaces.ants as ants

from nipype.interfaces.niftyreg import reg

from nipype.interfaces.niftyreg import regutils

from nipype.interfaces.ants.segmentation import DenoiseImage

from ..utils.utils_nodes import NodeParams, MapNodeParams
from ..utils.misc import parse_key

from ..nodes.prepare import average_align, FslOrient, reg_aladin_dirty
from ..nodes.extract_brain import T1xT2BET


def _create_reorient_pipeline(name="reorient_pipe",
                              new_dims=("x", "z", "-y")):
    """
    By kepkee:
    fslswapdim image_bad x z -y image_good
    fslorient -deleteorient image_good.nii.gz;
    fslorient -setqformcode 1 image_good.nii.gz
    """

    # creating pipeline
    reorient_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['image']),
        name='inputnode'
    )

    swap_dim = pe.Node(fsl.SwapDimensions(new_dims=new_dims),
                       name="swap_dim")
    reorient_pipe.connect(inputnode, 'image', swap_dim, 'in_file')

    deorient = pe.Node(FslOrient(main_option="deleteorient"),
                       name="deorient")
    reorient_pipe.connect(swap_dim, 'out_file', deorient, 'in_file')

    reorient = pe.Node(FslOrient(main_option="setqformcode", code=1),
                       name="reorient")
    reorient_pipe.connect(deorient, 'out_file', reorient, 'in_file')

    return reorient_pipe


def _create_mapnode_reorient_pipeline(name="reorient_pipe",
                                      new_dims=("x", "z", "-y")):
    """
    By kepkee (mapnode):
    fslswapdim image_bad x z -y image_good
    fslorient -deleteorient image_good.nii.gz;
    fslorient -setqformcode 1 image_good.nii.gz
    """
    # creating pipeline
    reorient_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_img']),
        name='inputnode'
    )

    swap_dim = pe.MapNode(fsl.SwapDimensions(new_dims=new_dims),
                          name="swap_dim", iterfield=["in_file"])
    reorient_pipe.connect(inputnode, 'list_img', swap_dim, 'in_file')

    deorient = pe.MapNode(FslOrient(main_option="deleteorient"),
                          name="deorient", iterfield=["in_file"])
    reorient_pipe.connect(swap_dim, 'out_file', deorient, 'in_file')

    reorient = pe.MapNode(FslOrient(main_option="setqformcode", code=1),
                          name="reorient", iterfield=["in_file"])
    reorient_pipe.connect(deorient, 'out_file', reorient, 'in_file')

    return reorient_pipe


def _create_prep_pipeline(params, name="prep_pipeline", node_suffix=""):
    """Description: hidden function for sequence of preprocessing

    Params:

    - reorient (new dims as a string (e.g. "x z -y"))
    - crop (see `ExtractROI <https://nipype.readthedocs.io/en/0.12.1/\
    interfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ \
    for arguments) - also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://nipype.readthedocs.io/en/0.12.\
    1/interfaces/generated/nipype.interfaces.ants.segmentation.html\
    #denoiseimage>`_ for arguments)) - also available \
    as :ref:`indiv_params <indiv_params>`
    """

    # init pipeline
    prep_pipeline = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['img', 'indiv_params']),
        name='inputnode')

    # Reorient if needed
    if "reorient" in params.keys():
        print("adding reorient pipeline")
        if "new_dims" in params["reorient"].keys():
            new_dims = tuple(params["reorient"]["new_dims"].split())
            print("using new dim")
        else:
            new_dims = ("x", "z", "-y")

        reorient_pipe = _create_reorient_pipeline(
            name="reorient_pipe", new_dims=new_dims)

        prep_pipeline.connect(inputnode, 'img',
                              reorient_pipe, 'inputnode.image')

    # cropping
    if 'crop'+node_suffix not in params.keys():
        print("Error, crop"+node_suffix+" could be found in params")
        print("Breaking")

        exit(0)

    crop = NodeParams(fsl.ExtractROI(),
                      params=parse_key(params, "crop"+node_suffix),
                      name='crop'+node_suffix)

    if "reorient" in params.keys():
        prep_pipeline.connect(reorient_pipe, 'reorient.out_file',
                              crop, 'in_file')
    else:
        prep_pipeline.connect(inputnode, 'img',
                              crop, 'in_file')
    prep_pipeline.connect(
        inputnode, ('indiv_params', parse_key, "crop"+node_suffix),
        crop, 'indiv_params')

    # N4 intensity normalization with parameters from json
    norm_intensity = NodeParams(ants.N4BiasFieldCorrection(),
                                params=parse_key(params, "norm_intensity"),
                                name='norm_intensity')

    prep_pipeline.connect(crop, 'roi_file',
                                norm_intensity, "input_image")
    prep_pipeline.connect(
        inputnode, ('indiv_params', parse_key, "norm_intensity"),
        norm_intensity, 'indiv_params')

    if "denoise" in params.keys():

        # denoise with Ants package
        # (taking into account whether reorient has been performed or not)
        denoise = NodeParams(interface=DenoiseImage(),
                             params=parse_key(params, "denoise"),
                             name="denoise")

        prep_pipeline.connect(norm_intensity, "output_image",
                              denoise, 'input_image')
        prep_pipeline.connect(
            inputnode, ('indiv_params', parse_key, "denoise"),
            denoise, 'indiv_params')

    # output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['prep_img']),
        name='outputnode')

    if "denoise" in params.keys():

        prep_pipeline.connect(denoise, 'output_image',
                              outputnode, "prep_img")

    else:

        prep_pipeline.connect(norm_intensity, "output_image",
                              outputnode, "prep_img")

    return prep_pipeline


def _create_mapnode_prep_pipeline(params, name="mapnode_prep_pipeline",
                                  node_suffix=""):

    """Description: hidden function for sequence of preprocessing (mapnode)

    Params: (Warning: **all arguments are in lists**)

    - reorient (new dims as a string (e.g. "x z -y"))
    - crop (see `ExtractROI <https://nipype.readthedocs.io/en/0.12.1/\
    interfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ \
    for arguments) - also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://nipype.readthedocs.io/en/0.12.\
    1/interfaces/generated/nipype.interfaces.ants.segmentation.html\
    #denoiseimage>`_ for arguments)) - also available \
    as :ref:`indiv_params <indiv_params>`
    """

    # init pipeline
    mapnode_prep_pipeline = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_img', 'indiv_params']),
        name='inputnode')

    # Reorient if needed
    if "reorient" in params.keys():

        if "new_dims" in params["reorient"].keys():
            new_dims = tuple(params["reorient"]["new_dims"].split())

        else:
            new_dims = ("x", "z", "-y")

            reorient_pipe = _create_mapnode_reorient_pipeline(
                name="reorient_pipe", new_dims=new_dims)

            mapnode_prep_pipeline.connect(inputnode, 'list_img',
                                          reorient_pipe, 'inputnode.list_img')

    if 'crop'+node_suffix not in params.keys():
        print("Error, crop"+node_suffix+" could be found in params")
        print("Breaking")

        exit(0)

    # cropping
    crop = MapNodeParams(fsl.ExtractROI(),
                         params=parse_key(params, "crop"+node_suffix),
                         name='crop', iterfield=["in_file", "args"])

    if "reorient" in params.keys():
        mapnode_prep_pipeline.connect(reorient_pipe, 'reorient.out_file',
                                      crop, 'in_file')
    else:
        mapnode_prep_pipeline.connect(inputnode, 'list_img',
                                      crop, 'in_file')
    mapnode_prep_pipeline.connect(
        inputnode, ('indiv_params', parse_key, "crop"+node_suffix),
        crop, 'indiv_params')

    # N4 intensity normalization with parameters from json
    norm_intensity = MapNodeParams(ants.N4BiasFieldCorrection(),
                                   params=parse_key(params, "norm_intensity"),
                                   name='norm_intensity',
                                   iterfield=['input_image'])

    mapnode_prep_pipeline.connect(crop, 'roi_file',
                                  norm_intensity, "input_image")
    mapnode_prep_pipeline.connect(
        inputnode, ("indiv_params", parse_key, "norm_intensity"),
        norm_intensity, 'indiv_params')

    if "denoise" in params.keys():

        # denoise with Ants package
        # (taking into account whether reorient has been performed or not)
        denoise = MapNodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise",
                                iterfield=["input_image"])

        mapnode_prep_pipeline.connect(norm_intensity, "output_image",
                                      denoise, "input_image")
        mapnode_prep_pipeline.connect(
            inputnode, ("indiv_params", parse_key, "denoise"),
            denoise, 'indiv_params')

    # output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['prep_list_img']),
        name='outputnode')

    if "denoise" in params.keys():

        mapnode_prep_pipeline.connect(denoise, "output_image",
                                      outputnode, "prep_list_img")

    else:

        mapnode_prep_pipeline.connect(norm_intensity, "output_image",
                                      outputnode, "prep_list_img")

    return mapnode_prep_pipeline


###############################################################################
# choices between the 3 main pipelines: "short", "long_single" et "long_multi"
###############################################################################
def create_short_preparation_pipe(params, params_template={},
                                  name="short_preparation_pipe"):
    """Description: short data preparation (average, reorient, crop/betcrop \
    and denoise)

    Processing steps;

    - T1s and T2s are averaged (by modality).
    - reorient (opt) makes the whole pipeline more neat)
    - Cropped (bet_crop include alignement, and mask generation)
    - denoising all images (no denoise_first)

    Params:

    - reorient (optional; new_dims as a string (e.g. "x z -y"))
    - bet_crop (:class:`T1xT2BET <macapype.nodes.extract_brain.T1xT2BET>`\
    ) or crop (see `ExtractRoi <https://nipype.readthedocs.io/en/0.12.1/inte\
    rfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ for \
    arguments). crop is also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://\
    nipype.readthedocs.io/en/0.12.1/interfaces/generated/nipype.interfaces\
    .ants.segmentation.html#denoiseimage>`_ for arguments))

    Inputs:

        inputnode:

            list_T1:
                T1 files (from BIDSDataGrabber)

            list_T2:
                T2 files

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:

            preproc_T1:
                preprocessed T1 file

            preproc_T2:
                preprocessed T2 file
    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['preproc_T1', 'preproc_T2']),
        name='outputnode')

    # avererge if multiple T2
    av_T2 = pe.Node(niu.Function(
        input_names=['list_img'],
        output_names=['avg_img'],
        function=average_align), name="av_T2")
    data_preparation_pipe.connect(inputnode, 'list_T2', av_T2, 'list_img')

    # avererge if multiple T1
    av_T1 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T1")
    data_preparation_pipe.connect(inputnode, 'list_T1', av_T1, 'list_img')

    if "crop_T1" in params.keys():
        print('crop_T1 is in params')

        assert "args" in params["crop_T1"].keys(), \
            "Error, args is not specified for crop node, breaking"

        # align avg T2 on avg T1
        align_T2_on_T1 = pe.Node(fsl.FLIRT(), name="align_T2_on_T1")
        align_T2_on_T1.inputs.dof = 6

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      align_T2_on_T1, 'reference')
        data_preparation_pipe.connect(av_T2, 'avg_img',
                                      align_T2_on_T1, 'in_file')

        # cropping
        # Crop bounding box for T1
        crop_T1 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop_T1'),
                             name='crop_T1')

        # Crop bounding box for T2
        crop_T2 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop_T1'),
                             name='crop_T2')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T1, 'indiv_params')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T2, 'indiv_params')

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      crop_T1, 'in_file')

        data_preparation_pipe.connect(align_T2_on_T1, "out_file",
                                      crop_T2, 'in_file')
    elif "bet_crop" in params.keys():

        # Brain extraction (unused) + Automated Cropping
        # default, if crop_T1 is undefined
        bet_crop = NodeParams(T1xT2BET(), params=params["bet_crop"],
                              name='bet_crop')

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      bet_crop, 't1_file')
        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      bet_crop, 't2_file')

    else:

        crop_aladin_T1 = NodeParams(reg.RegAladin(),
                                    params=parse_key(params, "crop_aladin_T1"),
                                    name='crop_aladin_T1')

        crop_aladin_T1.inputs.rig_only_flag = True
        crop_aladin_T1.inputs.nosym_flag = True
        crop_aladin_T1.inputs.ln_val = 12
        crop_aladin_T1.inputs.lp_val = 10
        crop_aladin_T1.inputs.smoo_r_val = 1.0

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      crop_aladin_T1, 'flo_file')

        crop_aladin_T1.inputs.ref_file = params_template["template_head"]

        # crop_z_T1
        crop_z_T1 = NodeParams(fsl.RobustFOV(),
                               params=parse_key(params, "crop_z_T1"),
                               name='crop_z_T1')

        data_preparation_pipe.connect(crop_aladin_T1, "res_file",
                                      crop_z_T1, 'in_file')

        # apply reg_resample to T2
        apply_crop_aladin_T2 = NodeParams(
            regutils.RegResample(),
            params=parse_key(params, "apply_crop_aladin_T2"),
            name='apply_crop_aladin_T2')

        data_preparation_pipe.connect(av_T2, 'avg_img',
                                      apply_crop_aladin_T2, 'flo_file')

        data_preparation_pipe.connect(crop_aladin_T1, 'aff_file',
                                      apply_crop_aladin_T2, 'trans_file')

        apply_crop_aladin_T2.inputs.ref_file = params_template["template_head"]

        if "align_crop_z_T2" in params.keys():

            # apply RobustFOV matrix as flirt
            align_crop_z_T2 = NodeParams(
                fsl.ApplyXFM(apply_xfm=True),
                params=parse_key(params, "align_crop_z_T2"),
                name="align_crop_z_T2")

            data_preparation_pipe.connect(
                crop_z_T1, 'out_roi', align_crop_z_T2, 'reference')

            data_preparation_pipe.connect(
                crop_z_T1, 'out_transform',
                align_crop_z_T2, 'in_matrix_file')

            data_preparation_pipe.connect(
                apply_crop_aladin_T2, 'out_file',
                align_crop_z_T2, 'in_file')

        else:

            # crop_z_T2
            crop_z_T2 = NodeParams(fsl.RobustFOV(),
                                   params=parse_key(params, "crop_z_T1"),
                                   name='crop_z_T2')

            data_preparation_pipe.connect(apply_crop_aladin_T2, 'out_file',
                                          crop_z_T2, 'in_file')

        # compute inv transfo
        inv_tranfo = NodeParams(
            regutils.RegTransform(),
            params=parse_key(params, "inv_transfo_aladin"),
            name='inv_tranfo')

        data_preparation_pipe.connect(crop_aladin_T1, 'aff_file',
                                      inv_tranfo, 'inv_aff_input')

    # denoise with Ants package
    if "denoise" in params.keys():

        denoise_T1 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T1")

        denoise_T2 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T2")
        # inputs
        if "crop_T1" in params.keys():
            data_preparation_pipe.connect(crop_T1, "roi_file",
                                          denoise_T1, 'input_image')

            data_preparation_pipe.connect(crop_T2, "roi_file",
                                          denoise_T2, 'input_image')

        elif "bet_crop" in params.keys():
            data_preparation_pipe.connect(bet_crop, "t1_cropped_file",
                                          denoise_T1, 'input_image')

            data_preparation_pipe.connect(bet_crop, "t2_cropped_file",
                                          denoise_T2, 'input_image')

        else:

            data_preparation_pipe.connect(crop_z_T1, "out_roi",
                                          denoise_T1, 'input_image')

            if "align_crop_z_T2" in params.keys():
                data_preparation_pipe.connect(align_crop_z_T2, "out_file",
                                              denoise_T2, 'input_image')
            else:
                data_preparation_pipe.connect(crop_z_T2, "out_roi",
                                              denoise_T2, 'input_image')

        # outputs
        data_preparation_pipe.connect(denoise_T1, 'output_image',
                                      outputnode, 'preproc_T1')

        data_preparation_pipe.connect(denoise_T2, 'output_image',
                                      outputnode, 'preproc_T2')
    else:

        if "crop_T1" in params.keys():
            data_preparation_pipe.connect(crop_T1, "roi_file",
                                          outputnode, 'preproc_T1')

            data_preparation_pipe.connect(crop_T2, "roi_file",
                                          outputnode, 'preproc_T2')

        elif "bet_crop" in params.keys():
            data_preparation_pipe.connect(bet_crop, "t1_cropped_file",
                                          outputnode, 'preproc_T1')

            data_preparation_pipe.connect(bet_crop, "t2_cropped_file",
                                          outputnode, 'preproc_T2')

        else:
            data_preparation_pipe.connect(crop_z_T1, "out_roi",
                                          outputnode, 'preproc_T1')

            if "align_crop_z_T2" in params.keys():
                data_preparation_pipe.connect(align_crop_z_T2, "out_file",
                                              outputnode, 'preproc_T2')
            else:
                data_preparation_pipe.connect(crop_z_T2, "out_roi",
                                              outputnode, 'preproc_T2')

    return data_preparation_pipe


def create_long_single_preparation_pipe(params,
                                        name="long_single_preparation_pipe"):
    """Description: Long data preparation, T1s and T2s are averaged
    (by modality) and then average imgs are processed.

    Processing Steps:

    - average imgs (by modality).
    - reorient (opt) makes the whole pipeline more neat)
    - start by denoising all input images (opt: denoise_first in json)
    - crop T1 and T2 based on individual parameters
    - running N4biascorrection on cropped T1 and T to help T2 and T1 alignment
    - finish by denoising all input images (opt: denoise in json))
    - Finally, align T2 to T1

    Params:

    - prep_T1 and prep_T2 (see :class:`_create_prep_pipeline <macapype.\
    pipelines.prepare._create_prep_pipeline>`)
    - align_T2_on_T1 (see `FLIRT <https://nipype.readthedocs.io/en/0.12.1/\
    interfaces/generated/nipype.interfaces.fsl.preprocess.html#flirt>`_ \
    for more arguments)

    Inputs:

        inputnode:
            list_T1: T1 files (from BIDSDataGrabber)
            list_T2: T2 files
            indiv_params (opt): dict with individuals parameters for some nodes

        arguments:
            params: dictionary of node sub-parameters (from a json file)
            name: pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:
            preproc_T1: preprocessed T1 file
            preproc_T2: preprocessed T2 file

    """

    # creating pipeline
    long_single_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # average if multiple T1
    av_T1 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T1")

    long_single_preparation_pipe.connect(
        inputnode, 'list_T1', av_T1, "list_img")

    # list_prep_pipeline for T1 list
    prep_T1_pipe = _create_prep_pipeline(
        params=parse_key(params, "prep_T1"),
        name="prep_T1", node_suffix="_T1")

    long_single_preparation_pipe.connect(
        av_T1, 'avg_img', prep_T1_pipe, "inputnode.img")

    long_single_preparation_pipe.connect(inputnode, 'indiv_params',
                                         prep_T1_pipe,
                                         "inputnode.indiv_params")

    # average if multiple T2
    av_T2 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T2")

    long_single_preparation_pipe.connect(
        inputnode, 'list_T2', av_T2, "list_img")

    # list_prep_pipeline for T2 list
    prep_T2_pipe = _create_prep_pipeline(
        params=parse_key(params, "prep_T2"),
        name="prep_T2", node_suffix="_T2")

    long_single_preparation_pipe.connect(
        av_T2, 'avg_img', prep_T2_pipe, "inputnode.img")

    long_single_preparation_pipe.connect(inputnode, 'indiv_params',
                                         prep_T2_pipe,
                                         "inputnode.indiv_params")

    # align T2 on T1 - mutualinfo works better for crossmodal biased images
    align_T2_on_T1 = NodeParams(fsl.FLIRT(),
                                params=parse_key(params, "align_T2_on_T1"),
                                name="align_T2_on_T1")

    long_single_preparation_pipe.connect(prep_T1_pipe, 'outputnode.prep_img',
                                         align_T2_on_T1, 'reference')
    long_single_preparation_pipe.connect(prep_T2_pipe, 'outputnode.prep_img',
                                         align_T2_on_T1, 'in_file')
    long_single_preparation_pipe.connect(
        inputnode, ("indiv_params", parse_key, "align_T2_on_T1"),
        align_T2_on_T1, "indiv_params")

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['preproc_T1', 'preproc_T2']),
        name='outputnode')

    long_single_preparation_pipe.connect(prep_T1_pipe, 'outputnode.prep_img',
                                         outputnode, 'preproc_T1')
    long_single_preparation_pipe.connect(align_T2_on_T1, 'out_file',
                                         outputnode, 'preproc_T2')

    return long_single_preparation_pipe


def create_long_multi_preparation_pipe(params,
                                       name="long_multi_preparation_pipe"):
    """Description: Long data preparation, each T1 and T2 is processed \
    independantly (mapnode).

    Processing steps:

    - reorient (opt: reorient) makes the whole pipeline more neat)
    - start by denoising all input images (opt: denoise_first in json)
    - crop T1 and T2 based on individual parameters
    - running N4biascorrection on cropped T1 and T to help T2 and T1 alignment
    - finish by denoising all input images (opt: denoise in json))
    - processed T1s and T2s are averaged (by modality)
    - Finally, align average T2 to average T1

    Params:

    - mapnode_prep_T1 and mapnodes_prep_T2 (see \
    :class:`_create_mapnode_prep_pipeline \
    <macapype.pipelines.prepare._create_mapnode_prep_pipeline>` for arguments)
    - align_T2_on_T1 (see `FLIRT <https://nipype.readthedocs.io/en/0.12.1/\
    interfaces/generated/nipype.interfaces.fsl.preprocess.html#flirt>`_ \
    for more arguments)

    Inputs:

        inputnode:

            list_T1:
                T1 files (from BIDSDataGrabber)

            list_T2:
                T2 files

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        :outputnode:

            preproc_T1:
                preprocessed T1 file

            preproc_T2:
                preprocessed T2 file

    """

    # creating pipeline
    long_multi_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['preproc_T1', 'preproc_T2']),
        name='outputnode')

    # mapnode_prep_pipeline for T1 list
    mapnode_prep_T1_pipe = _create_mapnode_prep_pipeline(
        params=parse_key(params, "prep_T1"),
        name="prep_T1", node_suffix="_T1")

    long_multi_preparation_pipe.connect(
        inputnode, 'list_T1', mapnode_prep_T1_pipe, "inputnode.list_img")

    long_multi_preparation_pipe.connect(
        inputnode, 'indiv_params',
        mapnode_prep_T1_pipe, "inputnode.indiv_params")

    # average if multiple T1
    av_T1 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T1")

    long_multi_preparation_pipe.connect(
        mapnode_prep_T1_pipe, 'outputnode.prep_list_img', av_T1, 'list_img')

    # mapnode_prep_pipeline for T2 list
    mapnode_prep_T2_pipe = _create_mapnode_prep_pipeline(
        params=parse_key(params, "prep_T2"),
        name="prep_T2", node_suffix="_T2")

    long_multi_preparation_pipe.connect(
        inputnode, 'list_T2', mapnode_prep_T2_pipe, "inputnode.list_img")

    long_multi_preparation_pipe.connect(
        inputnode, 'indiv_params',
        mapnode_prep_T2_pipe, "inputnode.indiv_params")

    # average if multiple T2
    av_T2 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T2")

    long_multi_preparation_pipe.connect(
        mapnode_prep_T2_pipe, 'outputnode.prep_list_img', av_T2, 'list_img')

    # align T2 on T1 - mutualinfo works better for crossmodal biased images
    align_T2_on_T1 = NodeParams(fsl.FLIRT(),
                                params=parse_key(params, "align_T2_on_T1"),
                                name="align_T2_on_T1")

    long_multi_preparation_pipe.connect(av_T1, 'avg_img',
                                        align_T2_on_T1, 'reference')
    long_multi_preparation_pipe.connect(av_T2, 'avg_img',
                                        align_T2_on_T1, 'in_file')

    # output node
    long_multi_preparation_pipe.connect(av_T1, 'avg_img',
                                        outputnode, 'preproc_T1')
    long_multi_preparation_pipe.connect(align_T2_on_T1, 'out_file',
                                        outputnode, 'preproc_T2')

    return long_multi_preparation_pipe


###############################################################################
# works with only one T1
def create_short_preparation_T1_pipe(params, params_template,
                                     name="short_preparation_T1_pipe"):
    """Description: T1 only short preparation

    Processing steps:

    - T1s and T2s are averaged (by modality).
    - reorient (opt: reorient) makes the whole pipeline more neat)
    - Cropped (bet_crop include alignement, and mask generation)
    - denoising all images (no denoise_first)

    Params:

    - reorient (optional; new_dims as a string (e.g. "x z -y"))
    - bet_crop (:class:`T1xT2BET <macapype.nodes.extract_brain.T1xT2BET>`\
    ) or crop (see `ExtractRoi <https://nipype.readthedocs.io/en/0.12.1/inte\
    rfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ for \
    arguments). crop is also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://\
    nipype.readthedocs.io/en/0.12.1/interfaces/generated/nipype.interfaces\
    .ants.segmentation.html#denoiseimage>`_ for arguments))

    Inputs:

        inputnode:

            list_T1:
                T1 files (from BIDSDataGrabber)

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:

            preproc_T1:
                preprocessed T1 file

    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'indiv_params']),
        name='inputnode'
    )

    # average if multiple T1
    av_T1 = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="av_T1")
    data_preparation_pipe.connect(inputnode, 'list_T1', av_T1, 'list_img')

    if "crop_T1" in params.keys():
        print('crop_T1 is in params')

        assert "args" in params["crop_T1"].keys(), \
            "Error, args is not specified for crop node, breaking"

        # cropping
        # Crop bounding box for T1
        crop_T1 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop'),
                             name='crop_T1')

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      crop_T1, 'in_file')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T1, 'indiv_params')

    elif "bet_crop" in params.keys():

        print('bet_crop is in params')

        # Brain extraction (unused) + Cropping
        bet_crop = NodeParams(T1xT2BET(), params=params["bet_crop"],
                              name='bet_crop')

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      bet_crop, 't1_file')
        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      bet_crop, 't2_file')

    else:
        crop_aladin_T1 = NodeParams(reg.RegAladin(),
                                    params=parse_key(params, "crop_aladin_T1"),
                                    name='crop_aladin_T1')

        crop_aladin_T1.inputs.rig_only_flag = True
        crop_aladin_T1.inputs.nosym_flag = True
        crop_aladin_T1.inputs.ln_val = 12
        crop_aladin_T1.inputs.lp_val = 10
        crop_aladin_T1.inputs.smoo_r_val = 1.0

        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      crop_aladin_T1, 'flo_file')

        crop_aladin_T1.inputs.ref_file = params_template["template_head"]

        # compute inv transfo
        inv_tranfo = NodeParams(
            regutils.RegTransform(),
            params=parse_key(params, "inv_transfo_aladin"),
            name='inv_tranfo')

        data_preparation_pipe.connect(crop_aladin_T1, 'aff_file',
                                      inv_tranfo, 'inv_aff_input')

        # crop_z_T1
        crop_z_T1 = NodeParams(fsl.RobustFOV(),
                               params=parse_key(params, "crop_z"),
                               name='crop_z_T1')

        data_preparation_pipe.connect(crop_aladin_T1, "res_file",
                                      crop_z_T1, 'in_file')

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['preproc_T1']),
        name='outputnode')

    if "denoise" in params.keys():

        # denoise with Ants package
        denoise_T1 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T1")

        # inputs
        if "bet_crop" in params.keys():
            data_preparation_pipe.connect(bet_crop, "t1_cropped_file",
                                          denoise_T1, 'input_image')

        elif "crop_T1" in params.keys():
            data_preparation_pipe.connect(crop_T1, "roi_file",
                                          denoise_T1, 'input_image')
        else:

            data_preparation_pipe.connect(crop_z_T1, "out_roi",
                                          denoise_T1, 'input_image')

        # outputs
        data_preparation_pipe.connect(denoise_T1, 'output_image',
                                      outputnode, 'preproc_T1')

    else:

        if "crop_T1" in params.keys():
            data_preparation_pipe.connect(crop_T1, "roi_file",
                                          outputnode, 'preproc_T1')
        elif "bet_crop" in params.keys():
            data_preparation_pipe.connect(bet_crop, "t1_cropped_file",
                                          outputnode, 'preproc_T1')

        else:
            data_preparation_pipe.connect(crop_z_T1, "out_roi",
                                          outputnode, 'preproc_T1')

    return data_preparation_pipe


###############################################################################
def create_short_preparation_FLAIR_pipe(params,
                                        name="short_preparation_FLAIR_pipe"):
    """Description: apply transfo on FLAIR (no reorient so far)

    Processing steps;

    - coreg FLAIR on T1
    - apply coreg on FLAIR

    Params:

    Inputs:

        inputnode:

            orig_T1:
                T1 files (from BIDSDataGrabber)

            FLAIR:
                FLAIR file

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:

            coreg_FLAIR:
                preprocessed FLAIR file

    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['orig_T1', 'FLAIR']),
        name='inputnode'
    )

    if "align_FLAIR_on_T1" in params.keys():
        print("in align_FLAIR_on_T1")

        # align FLAIR on avg T1
        align_FLAIR_on_T1 = NodeParams(
            fsl.FLIRT(),
            params=parse_key(params, "align_FLAIR_on_T1"),
            name="align_FLAIR_on_T1")

        data_preparation_pipe.connect(inputnode, 'orig_T1',
                                      align_FLAIR_on_T1, 'reference')

        data_preparation_pipe.connect(inputnode, 'FLAIR',
                                      align_FLAIR_on_T1, 'in_file')

    elif "reg_aladin_FLAIR_on_T1" in params.keys():
        print("in reg_aladin_FLAIR_on_T1")
        align_FLAIR_on_T1 = pe.Node(
            niu.Function(
                input_names=["reference", "in_file"],
                output_names=["out_file"],
                function=reg_aladin_dirty),
            name="reg_aladin_FLAIR_on_T1")

        # TODO
        # align_FLAIR_on_T1 = pe.Node(reg.RegAladin(),
        # name="reg_aladin_FLAIR_on_T1")

        data_preparation_pipe.connect(inputnode, 'orig_T1',
                                      align_FLAIR_on_T1, 'reference')

        data_preparation_pipe.connect(inputnode, 'FLAIR',
                                      align_FLAIR_on_T1, 'in_file')
    else:
        print("no align_FLAIR_on_T1 or reg_aladin_FLAIR_on_T1, breaking")
        exit(0)

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=['coreg_FLAIR']),
        name='outputnode')

    data_preparation_pipe.connect(align_FLAIR_on_T1, 'out_file',
                                  outputnode, 'coreg_FLAIR')

    return data_preparation_pipe


def create_short_preparation_MD_pipe(params,
                                     name="short_preparation_MD_pipe"):
    """Description: apply transfo on MD (no reorient so far)

    Processing steps;

    - init coreg b0mean on SS_T2
    - coreg b0mean on T2 using bbr and native_wm
    - apply coreg transfo on MD

    Params:

    Inputs:

        inputnode:

            orig_T2:
                T2 files (from BIDSDataGrabber)

            SS_T2:
                After Skull strip

            MD:
                MD file

            b0mean:
                B0 mean file

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:

            coreg_MD:
                preprocessed MD file with init

            coreg_better_MD:
                preprocessed MD file with init and flirt

    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['SS_T2', 'MD',
                                      'b0mean', 'native_wm_mask']),
        name='inputnode'
    )

    # init_align_b0mean_on_T2
    init_align_b0mean_on_T2 = NodeParams(
        fsl.FLIRT(), params=parse_key(params, "init_align_b0mean_on_T2"),
        name="init_align_b0mean_on_T2")

    data_preparation_pipe.connect(inputnode, 'SS_T2',
                                  init_align_b0mean_on_T2, 'reference')
    data_preparation_pipe.connect(inputnode, 'b0mean',
                                  init_align_b0mean_on_T2, 'in_file')

    # align_b0mean_on_T2
    align_b0mean_on_T2 = NodeParams(
        fsl.FLIRT(), params=parse_key(params, "align_b0mean_on_T2"),
        name="align_b0mean_on_T2")

    data_preparation_pipe.connect(inputnode, 'SS_T2',
                                  align_b0mean_on_T2, 'reference')
    data_preparation_pipe.connect(inputnode, 'b0mean',
                                  align_b0mean_on_T2, 'in_file')
    data_preparation_pipe.connect(inputnode, 'native_wm_mask',
                                  align_b0mean_on_T2, 'wm_seg')
    data_preparation_pipe.connect(init_align_b0mean_on_T2, 'out_matrix_file',
                                  align_b0mean_on_T2, 'in_matrix_file')

    # Apply transfo computed on b0 on MD (init)
    align_MD_on_T2_with_b0 = pe.Node(fsl.ApplyXFM(),
                                     name="align_MD_on_T2_with_b0")

    data_preparation_pipe.connect(inputnode, 'SS_T2',
                                  align_MD_on_T2_with_b0, 'reference')
    data_preparation_pipe.connect(inputnode, 'MD',
                                  align_MD_on_T2_with_b0, 'in_file')
    data_preparation_pipe.connect(init_align_b0mean_on_T2, 'out_matrix_file',
                                  align_MD_on_T2_with_b0, 'in_matrix_file')

    # Apply transfo computed on b0 on MD (second_flirt with GM)
    align_better_MD_on_T2_with_b0 = pe.Node(
        fsl.ApplyXFM(), name="align_better_MD_on_T2_with_b0")

    data_preparation_pipe.connect(inputnode, 'SS_T2',
                                  align_better_MD_on_T2_with_b0, 'reference')
    data_preparation_pipe.connect(inputnode, 'MD',
                                  align_better_MD_on_T2_with_b0, 'in_file')
    data_preparation_pipe.connect(align_b0mean_on_T2, 'out_matrix_file',
                                  align_better_MD_on_T2_with_b0,
                                  'in_matrix_file')

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=['coreg_MD', 'coreg_better_MD']),
        name='outputnode')

    data_preparation_pipe.connect(align_MD_on_T2_with_b0, 'out_file',
                                  outputnode, 'coreg_MD')

    data_preparation_pipe.connect(align_better_MD_on_T2_with_b0, 'out_file',
                                  outputnode, 'coreg_better_MD')

    return data_preparation_pipe



import os.path as op

import json
import pprint


def update_params(ssoft=[], subjects=None, sessions=None,
                  params_file=None, indiv_params_file=None):

    # params
    assert op.exists(params_file), "Error with file {}".format(
        params_file)

    params = json.load(open(params_file))

    extra_wf_name = ""
    indiv_params = {}

    # indiv_params
    if indiv_params_file is None:

        print("No indiv params where found, modifing pipeline to default")

        if "short_preparation_pipe" in params.keys():

            if "crop_T1" in params["short_preparation_pipe"].keys():
                print("Deleting crop_T1")
                del params["short_preparation_pipe"]["crop_T1"]

            print("New params after modification")
            pprint.pprint(params)

            extra_wf_name += "_crop_aladin"
        else:
            print('**** params modification not implemented \
                if no indiv_params ****')

    else:

        indiv_params = json.load(open(indiv_params_file))

        extra_wf_name += "_indiv_params"

        pprint.pprint(indiv_params)

        if "short_preparation_pipe" not in params.keys():

            print("short_preparation_pipe not found in params, \
                not modifying preparation pipe")

        else:

            prep_pipe = "short_preparation_pipe"
            count_all_sessions = 0
            count_T1_crops = 0
            count_long_crops = 0
            count_multi_long_crops = 0

            print("Indiv Params:", indiv_params_file)

            assert op.exists(indiv_params_file), "Error with file {}".format(
                indiv_params_file)

            if subjects is None or sessions is None:
                print("For whole BIDS dir, \
                    unable to assess if the indiv_params is correct")
                print("Running by default short_preparation_pipe and crop_T1")

                if "short_preparation_pipe" in params.keys():

                    if "crop_aladin_T1" in \
                            params["short_preparation_pipe"].keys():

                        print("Deleting crop_aladin_T1")
                        del params["short_preparation_pipe"]["crop_aladin_T1"]

                    if "crop_z_T1" in \
                            params["short_preparation_pipe"].keys():

                        print("Deleting crop_z_T1")
                        del params["short_preparation_pipe"]["crop_z_T1"]

                    print("Adding crop_T1")
                    params["short_preparation_pipe"]["crop_T1"] = \
                        {"args": "should be defined in indiv"}
            else:

                print("Will modify params if necessary, \
                    given specified subjects and sessions;\n")

                for sub in indiv_params.keys():

                    if sub.split('-')[1] not in subjects:
                        print('could not find subject {} in {}'.format(
                            sub.split('-')[1], subjects))
                        continue

                    for ses in indiv_params[sub].keys():

                        if ses.split('-')[1] not in sessions:

                            print('could not find session {} in {}'.format(
                                ses.split('-')[1], sessions))
                            continue

                        count_all_sessions += 1

                        indiv = indiv_params[sub][ses]

                        print(indiv.keys())

                        if "crop_T1" in indiv.keys():
                            count_T1_crops += 1

                            if "crop_T2" in indiv.keys() and 't1' not in ssoft:

                                count_long_crops += 1

                                crop_T1_args = indiv["crop_T1"]["args"]
                                crop_T2_args = indiv["crop_T2"]["args"]
                                if isinstance(crop_T1_args, list) and \
                                        isinstance(crop_T2_args, list):

                                    count_multi_long_crops += 1

                print("count_all_sessions {}".format(count_all_sessions))
                print("count_T1_crops {}".format(count_T1_crops))
                print("count_long_crops {}".format(count_long_crops))
                print("count_multi_long_crops {}".format(
                    count_multi_long_crops))

                if count_all_sessions:
                    if count_multi_long_crops == count_all_sessions:
                        print("**** Found list of crops for T1 and T2\
                            for all sub/ses in indiv -> \
                            long_multi_preparation_pipe")

                        prep_pipe = "long_multi_preparation_pipe"

                        extra_wf_name += "_long_multi"

                    elif count_long_crops == count_all_sessions:

                        print("**** Found crop for T1 and crop for T2\
                            for all sub/ses in indiv -> \
                            long_single_preparation_pipe")

                        prep_pipe = "long_single_preparation_pipe"

                        extra_wf_name += "_long_single"

                    elif count_T1_crops == count_all_sessions:

                        print("**** Found crop for T1 for all \
                            sub/ses in indiv \
                            -> keeping short_preparation_pipe")

                    else:
                        print("**** not all sub/ses have T1 and T2 crops ")
                        print("Error")
                        exit(0)

                if prep_pipe == "short_preparation_pipe":

                    if "crop_aladin_T1" in \
                            params["short_preparation_pipe"].keys():

                        print("Deleting crop_aladin_T1")
                        del params["short_preparation_pipe"]["crop_aladin_T1"]

                    if "crop_z_T1" in \
                            params["short_preparation_pipe"].keys():

                        print("Deleting crop_z_T1")
                        del params["short_preparation_pipe"]["crop_z_T1"]

                    print("Adding crop_T1")
                    params["short_preparation_pipe"]["crop_T1"] = \
                        {"args": "should be defined in indiv"}
                else:

                    params[prep_pipe] = {
                        "prep_T1": {
                            "crop_T1": {
                                "args": "should be defined in indiv"}},
                        "prep_T2": {
                            "crop_T2": {
                                "args": "should be defined in indiv"}},
                        "align_T2_on_T1": {"dof": 6, "cost": "normmi"}}

                    if "denoise" in params["short_preparation_pipe"].keys():
                        denoise = params["short_preparation_pipe"]["denoise"]

                        params[prep_pipe]["prep_T1"]["denoise"] = denoise
                        params[prep_pipe]["prep_T2"]["denoise"] = denoise

                    del params["short_preparation_pipe"]

    # prep for testing only preparation part
    if "prep" in ssoft:
        print("Found prep in soft")

        # ANTS
        if "brain_extraction_pipe" in params.keys():
            del params["brain_extraction_pipe"]
            print("Deleting brain_extraction_pipe")

        if "brain_segment_pipe" in params.keys():
            del params["brain_segment_pipe"]
            print("Deleting brain_segment_pipe")

        # SPM
        if "old_segment_pipe" in params.keys():
            del params["old_segment_pipe"]
            print("Deleting old_segment_pipe")

        if "debias" in params.keys():
            del params["debias"]
            print("Deleting debias")

        if "mask_from_seg_pipe" in params.keys():
            del params["mask_from_seg_pipe"]
            print("Deleting mask_from_seg_pipe")

    if "noseg" in ssoft:
        print("Found noseg in soft")

        # ANTS
        if "brain_segment_pipe" in params.keys():
            del params["brain_segment_pipe"]
            print("Deleting brain_segment_pipe")

        # SPM
        if "old_segment_pipe" in params.keys():
            del params["old_segment_pipe"]
            print("Deleting old_segment_pipe")

        if "mask_from_seg_pipe" in params.keys():
            del params["mask_from_seg_pipe"]
            print("Deleting mask_from_seg_pipe")

    print("After modif, running with params:")
    pprint.pprint(params)

    return params, indiv_params, extra_wf_name



#!/usr/bin/env python3
"""
    Non humain primates anatomical segmentation pipeline based ANTS

    Adapted in Nipype from an original pipelin of Kepkee Loh wrapped.

    Description
    --------------
    TODO :/

    Arguments
    -----------
    -data:
        Path to the BIDS directory that contain subjects' MRI data.

    -out:
        Nipype's processing directory.
        It's where all the outputs will be saved.

    -subjects:
        IDs list of subjects to process.

    -ses
        session (leave blank if None)

    -params
        json parameter file; leave blank if None

    Example
    ---------
    python segment_pnh.py -data [PATH_TO_BIDS] -out ../local_tests/ -subjects Elouk

    Requirements
    --------------
    This workflow use:
        - ANTS
        - AFNI
        - FSL
"""

# Authors : David Meunier (david.meunier@univ-amu.fr)
#           Bastien Cagna (bastien.cagna@univ-amu.fr)
#           Kepkee Loh (kepkee.loh@univ-amu.fr)
#           Julien Sein (julien.sein@univ-amu.fr)

import os
import os.path as op

import argparse
import json
import pprint

import nipype

import nipype.pipeline.engine as pe
import nipype.interfaces.utility as niu

import nipype.interfaces.fsl as fsl
fsl.FSLCommand.set_default_output_type('NIFTI_GZ')

from macapype.pipelines.full_pipelines import (
    create_full_spm_subpipes,
    create_full_ants_subpipes,
    create_full_T1_ants_subpipes,
    create_transfo_FLAIR_pipe,
    create_transfo_MD_pipe)

from macapype.utils.utils_bids import (create_datasource, create_datasource_indiv_params,
                                       create_datasink)

from macapype.utils.utils_tests import load_test_data, format_template

from macapype.utils.utils_nodes import node_output_exists
from macapype.utils.utils_params import update_params

from macapype.utils.misc import show_files, get_first_elem, parse_key

###############################################################################

def create_main_workflow(data_dir, process_dir, soft, species, subjects, sessions,
                         acquisitions, reconstructions, params_file,
                         indiv_params_file, mask_file, template_path,
                         template_files, nprocs, wf_name="macapype",
                         deriv=False, pad=False):

    # macapype_pipeline
    """ Set up the segmentatiopn pipeline based on ANTS

    Arguments
    ---------
    data_path: pathlike str
        Path to the BIDS directory that contains anatomical images

    out_path: pathlike str
        Path to the ouput directory (will be created if not alredy existing).
        Previous outputs maybe overwritten.

    soft: str
        Indicate which analysis should be launched; so for, only spm and ants
        are accepted; can be extended

    subjects: list of str (optional)
        Subject's IDs to match to BIDS specification (sub-[SUB1], sub-[SUB2]...)

    sessions: list of str (optional)
        Session's IDs to match to BIDS specification (ses-[SES1], ses-[SES2]...)

    acquisitions: list of str (optional)
        Acquisition name to match to BIDS specification (acq-[ACQ1]...)

    indiv_params_file: path to a JSON file
        JSON file that specify some parameters of the pipeline,
        unique for the subjects/sessions.

    params_file: path to a JSON file
        JSON file that specify some parameters of the pipeline.


    Returns
    -------
    workflow: nipype.pipeline.engine.Workflow


    """

    soft = soft.lower()

    ssoft = soft.split("_")

    new_ssoft = ssoft.copy()

    if 'test' in ssoft:
        new_ssoft.remove('test')

    if 'prep' in ssoft:
        new_ssoft.remove('prep')

    if 'noseg' in ssoft:
        new_ssoft.remove('noseg')

    if 't1' in ssoft:
        new_ssoft.remove('t1')

    if 'native' in ssoft:
        new_ssoft.remove('native')

    if 'template' in ssoft:
        new_ssoft.remove('template')

    soft = "_".join(new_ssoft)

    # formating args
    data_dir = op.abspath(data_dir)


    try:
        os.makedirs(process_dir)
    except OSError:
        print("process_dir {} already exists".format(process_dir))

    #params
    if params_file is None:

        # species
        if species is not None:

            species = species.lower()

            rep_species = {"marmoset":"marmo", "marmouset":"marmo", "chimpanzee":"chimp"}

            if species in list(rep_species.keys()):
                species = rep_species[species]

            list_species = ["macaque", "marmo", "baboon", "chimp"]

            assert species in list_species, \
                "Error, species {} should in the following list {}".format(
                    species, list_species)

            package_directory = op.dirname(op.abspath(__file__))

            params_file = "{}/params_segment_{}_{}.json".format(
                package_directory, species, soft)

        else:
            print("Error, no -params or no -species was found (one or the \
                other is mandatory)")
            exit(-1)

        print("Using default params file:", params_file)

        params, indiv_params, extra_wf_name = update_params(ssoft=ssoft, subjects=subjects, sessions=sessions,
                            params_file=params_file, indiv_params_file=indiv_params_file)

    else:

        # params
        assert op.exists(params_file), "Error with file {}".format(
            params_file)

        print("Using orig params file:", params_file)

        params = json.load(open(params_file))

        extra_wf_name = "_orig"
        indiv_params = {}

        # indiv_params
        if indiv_params_file is not None:

            assert op.exists(indiv_params_file), "Error with file {}".format(
                indiv_params_file)
            indiv_params = json.load(open(indiv_params_file))

    wf_name += extra_wf_name

    # soft
    wf_name += "_{}".format(soft)

    if 't1' in ssoft:
        wf_name += "_t1"

    if mask_file is not None:
        wf_name += "_mask"

    assert "spm" in ssoft or "spm12" in ssoft or "ants" in ssoft, \
        "error with {}, should be among [spm12, spm, ants]".format(ssoft)

    # adding forced space
    if "spm" in ssoft or "spm12" in ssoft:
        if 'native' in ssoft:
            wf_name += "_native"

    elif "ants" in ssoft:
        if "template" in ssoft:
            wf_name += "_template"

    # params_template
    if template_path is not None:
        print(template_files)

        params_template = {}

        assert len(template_files) > 1, "Error, template_files unspecified {}".format(template_files)

        template_head = os.path.join(template_path,template_files[0])
        assert os.path.exists(template_head), "Could not find template_head {}".format(template_head)
        params_template["template_head"] = template_head

        template_brain = os.path.join(template_path,template_files[1])
        assert os.path.exists(template_brain), "Could not find template_brain {}".format(template_brain)
        params_template["template_brain"] = template_brain

        if len(template_files) == 3:

            template_seg = os.path.join(template_path,template_files[2])
            assert os.path.exists(template_seg), "Could not find template_seg {}".format(template_seg)
            params_template["template_seg"] = template_seg

        elif len(template_files) == 5:

            template_gm = os.path.join(template_path,template_files[2])
            assert os.path.exists(template_gm), "Could not find template_gm {}".format(template_gm)
            params_template["template_gm"] = template_gm

            template_wm = os.path.join(template_path,template_files[3])
            assert os.path.exists(template_wm), "Could not find template_wm {}".format(template_wm)
            params_template["template_wm"] = template_wm

            template_csf = os.path.join(template_path,template_files[4])
            assert os.path.exists(template_csf), "Could not find template_csf {}".format(template_csf)
            params_template["template_csf"] = template_csf

        else:
            print("Unknown template_files format, should be 3 or 5 files")
            exit(-1)

    else:
        ### use template from params
        assert ("general" in params.keys() and \
            "template_name" in params["general"].keys()), \
                "Error, the params.json should contains a general/template_name"

        template_name = params["general"]["template_name"]

        if "general" in params.keys() and "my_path" in params["general"].keys():
            my_path = params["general"]["my_path"]
        else:
            my_path = ""

        template_dir = load_test_data(template_name, path_to = my_path)
        params_template = format_template(template_dir, template_name)

        if "template_aladin_name" in params["general"].keys():

            template_aladin_name = params["general"]["template_aladin_name"]
            template_aladin_dir = load_test_data(template_aladin_name, path_to = my_path)
            params_template_aladin = format_template(template_aladin_dir, template_aladin_name)

        else:
            params_template_aladin = params_template




    print (params_template)

    # main_workflow
    main_workflow = pe.Workflow(name= wf_name)

    main_workflow.base_dir = process_dir

    # which soft is used
    if "spm" in ssoft or "spm12" in ssoft:
        if 'native' in ssoft:
            space='native'

        else:
            space='template'

        segment_pnh_pipe = create_full_spm_subpipes(
            params_template=params_template,
            params_template_aladin=params_template_aladin,
            params=params, pad=pad, space=space)

    elif "ants" in ssoft:
        if "template" in ssoft:
            space="template"

        else:
            space="native"

        if "t1" in ssoft:
            segment_pnh_pipe = create_full_T1_ants_subpipes(
                params_template=params_template,
                params_template_aladin=params_template_aladin,
                params=params, space=space, pad=pad)
        else:
            segment_pnh_pipe = create_full_ants_subpipes(
                params_template=params_template,
                params_template_aladin=params_template_aladin,
                params=params, mask_file=mask_file, space=space, pad=pad)

    # list of all required outputs
    output_query = {}

    # T1 (mandatory, always added)
    # T2 is optional, if "_T1" is added in the -soft arg
    if 't1' in ssoft:
        output_query['T1'] = {
            "datatype": "anat", "suffix": "T1w",
            "extension": ["nii", ".nii.gz"]}

    else:
        output_query['T1'] = {
            "datatype": "anat", "suffix": "T1w",
            "extension": ["nii", ".nii.gz"]}

        output_query['T2'] = {
            "datatype": "anat", "suffix": "T2w",
            "extension": ["nii", ".nii.gz"]}

    # FLAIR is optional, if "_FLAIR" is added in the -soft arg
    if 'flair' in ssoft:
        output_query['FLAIR'] = {
            "datatype": "anat", "suffix": "FLAIR",
            "extension": ["nii", ".nii.gz"]}

    # MD and b0mean are optional, if "_MD" is added in the -soft arg
    if 'md' in ssoft:
        output_query['MD'] =  {
            "datatype": "dwi", "acquisition": "MD", "suffix": "dwi",
            "extension": ["nii", ".nii.gz"]}

        output_query['b0mean'] = {
            "datatype": "dwi", "acquisition": "b0mean", "suffix": "dwi",
            "extension": ["nii", ".nii.gz"]}

    # indiv_params
    if indiv_params:
        datasource = create_datasource_indiv_params(
            output_query, data_dir, indiv_params, subjects, sessions,
            acquisitions, reconstructions)

        main_workflow.connect(datasource, "indiv_params",
                              segment_pnh_pipe,'inputnode.indiv_params')
    else:
        datasource = create_datasource(
            output_query, data_dir, subjects,  sessions, acquisitions,
            reconstructions)

    main_workflow.connect(datasource, 'T1',
                          segment_pnh_pipe, 'inputnode.list_T1')

    if not "t1" in ssoft:
        main_workflow.connect(datasource, 'T2',
                              segment_pnh_pipe, 'inputnode.list_T2')
    elif "t1" in ssoft and "spm" in ssoft:
        # cheating using T2 as T1
        main_workflow.connect(datasource, 'T1',
                              segment_pnh_pipe, 'inputnode.list_T2')

    if "flair" in ssoft:


        if "transfo_FLAIR_pipe" in params.keys():
            print("Found transfo_FLAIR_pipe")

        transfo_FLAIR_pipe = create_transfo_FLAIR_pipe(params=parse_key(params, "transfo_FLAIR_pipe"),
                                                       params_template=params_template)

        if "t1" in ssoft:
            main_workflow.connect(segment_pnh_pipe, "short_preparation_pipe.outputnode.preproc_T1",
                                transfo_FLAIR_pipe, 'inputnode.orig_T1')

        else:
            main_workflow.connect(segment_pnh_pipe, "debias.t1_debiased_file",
                                transfo_FLAIR_pipe, 'inputnode.orig_T1')


        main_workflow.connect(segment_pnh_pipe, "reg.transfo_file",
                              transfo_FLAIR_pipe, 'inputnode.lin_transfo_file')

        main_workflow.connect(datasource, ('FLAIR', get_first_elem),
                              transfo_FLAIR_pipe, 'inputnode.FLAIR')

    if 'md' in ssoft:

        if "transfo_MD_pipe" in params.keys():
            print("Found transfo_MD_pipe")

        transfo_MD_pipe = create_transfo_MD_pipe(params=parse_key(params, "transfo_MD_pipe"),
                                                 params_template=params_template)

        main_workflow.connect(segment_pnh_pipe,
                                "old_segment_pipe.outputnode.threshold_wm",
                                transfo_MD_pipe, 'inputnode.threshold_wm')

        main_workflow.connect(datasource, ('MD', get_first_elem),
                                transfo_MD_pipe, 'inputnode.MD')

        main_workflow.connect(datasource, ('b0mean', get_first_elem),
                                transfo_MD_pipe, 'inputnode.b0mean')

        main_workflow.connect(segment_pnh_pipe, "debias.t1_debiased_file",
                            transfo_MD_pipe, 'inputnode.orig_T1')

        main_workflow.connect(segment_pnh_pipe, "debias.t2_debiased_brain_file",
                            transfo_MD_pipe, 'inputnode.SS_T2')

        main_workflow.connect(segment_pnh_pipe, "reg.transfo_file",
                            transfo_MD_pipe, 'inputnode.lin_transfo_file')

        main_workflow.connect(segment_pnh_pipe, "reg.inv_transfo_file",
                            transfo_MD_pipe, 'inputnode.inv_lin_transfo_file')

    if deriv:

        datasink_name = os.path.join("derivatives", wf_name)

        if "regex_subs" in params.keys():
            params_regex_subs = params["regex_subs"]
        else:
            params_regex_subs={}

        if "subs" in params.keys():
            params_subs = params["rsubs"]
        else:
            params_subs={}

        print (datasource.iterables)

        datasink = create_datasink(iterables=datasource.iterables,
                                   name=datasink_name,
                                   params_subs=params_subs,
                                   params_regex_subs=params_regex_subs)

        datasink.inputs.base_directory = process_dir

        if "brain_extraction_pipe" in params.keys():

            ### rename brain_mask
            rename_brain_mask = pe.Node(niu.Rename(), name = "rename_brain_mask")
            rename_brain_mask.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-brain_mask"
            rename_brain_mask.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_brain_mask.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.brain_mask',
                rename_brain_mask, 'in_file')

            main_workflow.connect(
                rename_brain_mask, 'out_file',
                datasink, '@brain_mask')

        elif "debias" in params.keys():

            ### rename brain_mask
            rename_brain_mask = pe.Node(niu.Rename(), name = "rename_brain_mask")
            rename_brain_mask.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-brain_mask"
            rename_brain_mask.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_brain_mask.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.brain_mask',
                rename_brain_mask, 'in_file')

            main_workflow.connect(
                rename_brain_mask, 'out_file',
                datasink, '@brain_mask')

            ### rename debiased_T1
            rename_debiased_T1 = pe.Node(niu.Rename(), name = "rename_debiased_T1")
            rename_debiased_T1.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-debiased_T1w"
            rename_debiased_T1.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_debiased_T1.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.debiased_T1',
                rename_debiased_T1, 'in_file')

            main_workflow.connect(
                rename_debiased_T1, 'out_file',
                datasink, '@debiased_T1')

            ### rename debiased_brain
            rename_debiased_brain = pe.Node(niu.Rename(), name = "rename_debiased_brain")
            rename_debiased_brain.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-debiased_desc-brain_T1w"
            rename_debiased_brain.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_debiased_brain.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.debiased_brain',
                rename_debiased_brain, 'in_file')

            main_workflow.connect(
                rename_debiased_brain, 'out_file',
                datasink, '@debiased_brain')

        if "brain_segment_pipe" in params.keys():

            ### rename debiased_brain
            rename_debiased_brain = pe.Node(niu.Rename(), name = "rename_debiased_brain")
            rename_debiased_brain.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-debiased_desc-brain_T1w"
            rename_debiased_brain.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_debiased_brain.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.debiased_brain',
                rename_debiased_brain, 'in_file')

            main_workflow.connect(
                rename_debiased_brain, 'out_file',
                datasink, '@debiased_brain')

            ### rename debiased_T1
            rename_debiased_T1 = pe.Node(niu.Rename(), name = "rename_debiased_T1")
            rename_debiased_T1.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-native_desc-debiased_T1w"
            rename_debiased_T1.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_debiased_T1.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.debiased_T1',
                rename_debiased_T1, 'in_file')

            main_workflow.connect(
                rename_debiased_T1, 'out_file',
                datasink, '@debiased_T1')

            ### rename segmented_brain_mask
            rename_segmented_brain_mask = pe.Node(niu.Rename(), name = "rename_segmented_brain_mask")
            rename_segmented_brain_mask.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_desc-brain_dseg".format(space)
            rename_segmented_brain_mask.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_segmented_brain_mask.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.segmented_brain_mask',
                rename_segmented_brain_mask, 'in_file')

            main_workflow.connect(
                rename_segmented_brain_mask, 'out_file',
                datasink, '@segmented_brain_mask')

            ### rename prob_wm
            print("Renaming prob_wm file")
            rename_prob_wm = pe.Node(niu.Rename(), name = "rename_prob_wm")
            rename_prob_wm.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-WM_probseg".format(space)
            rename_prob_wm.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_wm.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_wm',
                rename_prob_wm, 'in_file')

            main_workflow.connect(
                rename_prob_wm, ('out_file', show_files),
                datasink, '@prob_wm')

            ### rename prob_gm
            print("Renaming prob_gm file")
            rename_prob_gm = pe.Node(niu.Rename(), name = "rename_prob_gm")
            rename_prob_gm.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-GM_probseg".format(space)
            rename_prob_gm.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_gm.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_gm',
                rename_prob_gm, 'in_file')

            main_workflow.connect(
                rename_prob_gm, ('out_file', show_files),
                datasink, '@prob_gm')

            ### rename prob_csf
            print("Renaming prob_csf file")
            rename_prob_csf = pe.Node(niu.Rename(), name = "rename_prob_csf")
            rename_prob_csf.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-CSF_probseg".format(space)
            rename_prob_csf.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_csf.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_csf',
                rename_prob_csf, 'in_file')

            main_workflow.connect(
                rename_prob_csf, ('out_file', show_files),
                datasink, '@prob_csf')

            # rename 5tt
            if "export_5tt_pipe" in params["brain_segment_pipe"]:
                print("Renaming 5tt file")

                rename_gen_5tt = pe.Node(niu.Rename(), name = "rename_gen_5tt")
                rename_gen_5tt.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_desc-5tt_dseg".format(space)
                rename_gen_5tt.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
                rename_gen_5tt.inputs.keep_ext = True

                main_workflow.connect(
                    segment_pnh_pipe, 'outputnode.gen_5tt',
                    rename_gen_5tt, 'in_file')

                main_workflow.connect(
                    rename_gen_5tt, 'out_file',
                    datasink, '@gen_5tt')

        elif "old_segment_pipe" in params.keys():

            ### rename prob_wm
            print("Renaming prob_wm file")
            rename_prob_wm = pe.Node(niu.Rename(), name = "rename_prob_wm")
            rename_prob_wm.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-WM_probseg".format(space)
            rename_prob_wm.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_wm.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_wm',
                rename_prob_wm, 'in_file')

            main_workflow.connect(
                rename_prob_wm, ('out_file', show_files),
                datasink, '@prob_wm')

            ### rename prob_gm
            print("Renaming prob_gm file")
            rename_prob_gm = pe.Node(niu.Rename(), name = "rename_prob_gm")
            rename_prob_gm.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-GM_probseg".format(space)
            rename_prob_gm.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_gm.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_gm',
                rename_prob_gm, 'in_file')

            main_workflow.connect(
                rename_prob_gm, ('out_file', show_files),
                datasink, '@prob_gm')

            ### rename prob_csf
            print("Renaming prob_csf file")
            rename_prob_csf = pe.Node(niu.Rename(), name = "rename_prob_csf")
            rename_prob_csf.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_label-CSF_probseg".format(space)
            rename_prob_csf.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_prob_csf.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.prob_csf',
                rename_prob_csf, 'in_file')

            main_workflow.connect(
                rename_prob_csf, ('out_file', show_files),
                datasink, '@prob_csf')

            ## rename 5tt
            #if "export_5tt_pipe" in params["brain_segment_pipe"]:
                #print("Renaming 5tt file")

                #rename_gen_5tt = pe.Node(niu.Rename(), name = "rename_gen_5tt")
                #rename_gen_5tt.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_desc-5tt_dseg".format(space)
                #rename_gen_5tt.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
                #rename_gen_5tt.inputs.keep_ext = True

                #main_workflow.connect(
                    #segment_pnh_pipe, 'outputnode.gen_5tt',
                    #rename_gen_5tt, 'in_file')

                #main_workflow.connect(
                    #rename_gen_5tt, 'out_file',
                    #datasink, '@gen_5tt')



        if "mask_from_seg_pipe" in params.keys():

            ### rename segmented_brain_mask
            rename_segmented_brain_mask = pe.Node(niu.Rename(), name = "rename_segmented_brain_mask")
            rename_segmented_brain_mask.inputs.format_string = "sub-%(sub)s_ses-%(ses)s_space-{}_desc-brain_dseg".format(space)
            rename_segmented_brain_mask.inputs.parse_string = r"sub-(?P<sub>\w*)_ses-(?P<ses>\w*)_.*"
            rename_segmented_brain_mask.inputs.keep_ext = True

            main_workflow.connect(
                segment_pnh_pipe, 'outputnode.segmented_brain_mask',
                rename_segmented_brain_mask, 'in_file')

            main_workflow.connect(
                rename_segmented_brain_mask, 'out_file',
                datasink, '@segmented_brain_mask')

        if 'flair' in ssoft :

            main_workflow.connect(
                transfo_FLAIR_pipe, 'outputnode.norm_FLAIR',
                datasink, '@norm_flair')

    main_workflow.write_graph(graph2use="colored")
    main_workflow.config['execution'] = {'remove_unnecessary_outputs': 'false'}

    # saving real params.json
    real_params_file = op.join(process_dir, wf_name, "real_params.json")
    with open(real_params_file, 'w+') as fp:
        json.dump(params, fp)

    if nprocs is None:
        nprocs = 4

    if not "test" in ssoft:
        if "seq" in ssoft or nprocs==0:
            main_workflow.run()
        else:
            main_workflow.run(plugin='MultiProc',
                              plugin_args={'n_procs' : nprocs})
def main():

    # Command line parser
    parser = argparse.ArgumentParser(
        description="PNH segmentation pipeline")

    parser.add_argument("-data", dest="data", type=str, required=True,
                        help="Directory containing MRI data (BIDS)")
    parser.add_argument("-out", dest="out", type=str, #nargs='+',
                        help="Output dir", required=True)
    parser.add_argument("-soft", dest="soft", type=str,
                        help="Sofware of analysis (SPM or ANTS are defined)",
                        required=True)
    parser.add_argument("-species", dest="species", type=str,
                        help="Type of PNH to process",
                        required=False)
    parser.add_argument("-subjects", "-sub", dest="sub",
                        type=str, nargs='+', help="Subjects", required=False)
    parser.add_argument("-sessions", "-ses", dest="ses",
                        type=str, nargs='+', help="Sessions", required=False)
    parser.add_argument("-acquisitions", "-acq", dest="acq", type=str,
                        nargs='+', default=None, help="Acquisitions")
    parser.add_argument("-records", "-rec", dest="rec", type=str, nargs='+',
                        default=None, help="Records")
    parser.add_argument("-params", dest="params_file", type=str,
                        help="Parameters json file", required=False)
    parser.add_argument("-indiv_params", "-indiv", dest="indiv_params_file",
                        type=str, help="Individual parameters json file",
                        required=False)
    parser.add_argument("-mask", dest="mask_file", type=str,
                        help="precomputed mask file", required=False)

    parser.add_argument("-template_path", dest="template_path", type=str,
                        help="specifies user-based template path",
                        required=False)

    parser.add_argument("-template_files", dest="template_files", type=str,
                        nargs="+", help="specifies user-based template files \
                            (3 or 5 are possible options)",
                        required=False)

    parser.add_argument("-nprocs", dest="nprocs", type=int,
                        help="number of processes to allocate", required=False)

    parser.add_argument("-deriv", dest="deriv", action='store_true',
                        help="output derivatives in BIDS orig directory",
                        required=False)

    parser.add_argument("-pad", dest="pad", action='store_true',
                        help="padding mask and seg_mask",
                        required=False)


    args = parser.parse_args()

    # main_workflow
    print("Initialising the pipeline...")
    create_main_workflow(
        data_dir=args.data,
        soft=args.soft,
        process_dir=args.out,
        species=args.species,
        subjects=args.sub,
        sessions=args.ses,
        acquisitions=args.acq,
        reconstructions=args.rec,
        params_file=args.params_file,
        indiv_params_file=args.indiv_params_file,
        mask_file=args.mask_file,
        template_path=args.template_path,
        template_files=args.template_files,
        nprocs=args.nprocs,
        deriv=args.deriv,
        pad=args.pad)

if __name__ == '__main__':
    main()



import shutil

import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

import nipype.interfaces.fsl as fsl
import nipype.interfaces.afni as afni
import nipype.interfaces.ants as ants

from ..utils.misc import get_elem
from ..utils.utils_nodes import NodeParams, parse_key

from ..nodes.register import (interative_flirt, NMTSubjectAlign,
                              NMTSubjectAlign2, NwarpApplyPriors,
                              animal_warper)


def create_iterative_register_pipe(
        template_file, template_brain_file, template_mask_file, gm_prob_file,
        wm_prob_file, csf_prob_file, n_iter, name="register_pipe"):
    """
    Registration of template (NMT or other) according to Regis:

    - The iterative FLIRT is between NMT_SS and subject's anat after a quick
    skull-stripping but the anat is more and more refined to corresponds to the
    brain

    - there is also a FNIRT done once, for comparison of the quality of the
    template on the subject's brain

    Not used anymore: corresponds to the IterREGBET provided by Regis in bash
    and wrapped node IterREGBET in nodes/register.py

    #TODO: test if gives the same results as IterREGBET
    """
    register_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_file_BET', 'anat_file']),
        name='inputnode')

    # register node
    register = pe.Node(
        niu.Function(input_names=["anat_file", "anat_file_BET",
                                  "template_brain_file",
                                  "template_mask_file", 'n_iter'],
                     output_names=["anat_file_brain", "template_to_anat_file"],
                     function=interative_flirt),
        name="register")

    register.inputs.template_brain_file = template_brain_file
    register.inputs.template_mask_file = template_mask_file
    register.inputs.n_iter = n_iter

    register_pipe.connect(inputnode, 'anat_file', register, 'anat_file')
    register_pipe.connect(inputnode, 'anat_file_BET',
                          register, "anat_file_BET")

    # apply transfo over the 3 tissues:
    # gm
    register_gm = pe.Node(fsl.ApplyXFM(), name="register_gm")

    register_gm.inputs.in_file = gm_prob_file
    register_gm.inputs.apply_xfm = True
    register_gm.inputs.interp = "nearestneighbour"
    register_gm.inputs.output_type = "NIFTI"  # for SPM segment

    register_pipe.connect(register, 'anat_file_brain',
                          register_gm, 'reference')
    register_pipe.connect(register, 'template_to_anat_file',
                          register_gm, "in_matrix_file")

    # wm
    register_wm = pe.Node(fsl.ApplyXFM(), name="register_wm")

    register_wm.inputs.in_file = wm_prob_file
    register_wm.inputs.apply_xfm = True
    register_wm.inputs.interp = "nearestneighbour"
    register_wm.inputs.output_type = "NIFTI"  # for SPM segment

    register_pipe.connect(register, 'anat_file_brain',
                          register_wm, 'reference')
    register_pipe.connect(register, 'template_to_anat_file',
                          register_wm, "in_matrix_file")

    # csf
    register_csf = pe.Node(fsl.ApplyXFM(), name="register_csf")

    register_csf.inputs.in_file = csf_prob_file
    register_csf.inputs.apply_xfm = True
    register_csf.inputs.interp = "nearestneighbour"
    register_csf.inputs.output_type = "NIFTI"  # for SPM segment

    register_pipe.connect(register, 'anat_file_brain',
                          register_csf, 'reference')
    register_pipe.connect(register, 'template_to_anat_file',
                          register_csf, "in_matrix_file")

    def return_list(file1, file2, file3):
        return [file1, file2, file3]

    # merge 3 outputs to a list (...)
    merge_3_files = pe.Node(
        niu.Function(input_names=["file1", "file2", "file3"],
                     output_names=["list3files"],
                     function=return_list),
        name="merge_3_files")

    register_pipe.connect(register_gm, 'out_file', merge_3_files, "file1")
    register_pipe.connect(register_wm, 'out_file', merge_3_files, "file2")
    register_pipe.connect(register_csf, 'out_file', merge_3_files, "file3")

    # same with non linear
    # non linear register between anat and head
    # (FNIRT work directly on head?
    # I thought FLIRT was only skull-stripped brain, is it different? )

    nl_register = pe.Node(fsl.FNIRT(), name="nl_register")
    nl_register.inputs.in_file = template_file

    register_pipe.connect(inputnode, 'anat_file', nl_register, 'ref_file')
    register_pipe.connect(register, 'template_to_anat_file',
                          nl_register, 'affine_file')

    # apply non linear warp to NMT_SS
    nl_apply = pe.Node(fsl.ApplyWarp(), name="nl_apply")
    nl_apply.inputs.in_file = template_brain_file
    register_pipe.connect(inputnode, 'anat_file', nl_apply, 'ref_file')
    register_pipe.connect(nl_register, 'fieldcoeff_file',
                          nl_apply, 'field_file')  # iout from fnirt

    return register_pipe


###############################################################################
def create_register_NMT_pipe(params_template, params={},
                             name="register_NMT_pipe"):
    """Description: Register template to anat with the script NMT_subject_align

    Processing steps:

    - Bias correction (norm_intensity)
    - Deoblique (Refit with deoblique option)
    - NMT_subject_align (see :class:`NMTSubjectAlign \
    <macapype.nodes.register.NMTSubjectAlign>` and :class:`NMTSubjectAlign2 \
    <macapype.nodes.register.NMTSubjectAlign2>` for explanations)
    - apply it to tissues list_priors (NwarpApplyPriors and Allineate)

    Params:
        - norm_intensity (see `N4BiasFieldCorrection <https://\
        nipype.readthedocs.io/en/0.12.1/interfaces/generated/nipype.interfaces\
        .ants.segmentation.html#n4biasfieldcorrection>`_ for arguments)) - \
        also available as :ref:`indiv_params <indiv_params>`

    Inputs:

        inputnode:

            T1:
                T1 file name

            indiv_params:
                dict with individuals parameters for some nodes

        arguments:

            params_template:
                dictionary of info about template

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "register_NMT_pipe")

    Outputs:

        norm_intensity.output_image:
            filled mask after erode

        align_seg_csf.out_file:
            csf template tissue in subject space

        align_seg_gm.out_file:
            grey matter template tissue in subject space

        align_seg_wm.out_file:
            white matter template tissue in subject space
    """

    register_NMT_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['T1', 'indiv_params']),
        name='inputnode')

    if "norm_intensity" in params.keys():

        # N4 intensity normalization over brain
        norm_intensity = NodeParams(ants.N4BiasFieldCorrection(),
                                    params=parse_key(params, "norm_intensity"),
                                    name='norm_intensity')

        register_NMT_pipe.connect(inputnode, 'T1',
                                  norm_intensity, "input_image")

        register_NMT_pipe.connect(
            inputnode, ('indiv_params', parse_key, "norm_intensity"),
            norm_intensity, "indiv_params")

    # deoblique (seems mandatory from NMTSubjectAlign)
    deoblique = pe.Node(afni.Refit(deoblique=True), name="deoblique")

    if "norm_intensity" in params.keys():

        register_NMT_pipe.connect(norm_intensity, 'output_image',
                                  deoblique, "in_file")

    else:

        register_NMT_pipe.connect(inputnode, 'T1',
                                  deoblique, "in_file")

    print("which @animal_warper: ", shutil.which("@animal_warper"))

    if "NMT_subject_align" in params.keys():

        if "NMT_version" in params["NMT_subject_align"].keys():
            NMT_version = params["NMT_subject_align"]['NMT_version']
        else:
            print("Warning, for NMT_subject_align, NMT_version is required")
            NMT_version = "v1.3"

        print("*** NMT_version: {}".format(NMT_version))

        print("running NMT_subject_align with version {}".format(NMT_version))

        if NMT_version == "v1.2":
            # align subj to nmt
            NMT_subject_align = NodeParams(
                NMTSubjectAlign(),
                params=parse_key(params, "NMT_subject_align"),
                name='NMT_subject_align')

        elif NMT_version == "v1.3" or NMT_version == "v2.0":
            # align subj to nmt
            NMT_subject_align = NodeParams(
                NMTSubjectAlign2(),
                params=parse_key(params, "NMT_subject_align"),
                name='NMT_subject_align')

        else:
            print("NMT_version {} is not implemented, breaking".format(
                NMT_version))

            exit()

    elif shutil.which("@animal_warper") is not None:
        # TODO AnimalWarper()

        print("running @animal_warper with version {}".format(
            shutil.which("@animal_warper")))

        NMT_subject_align = pe.Node(
            niu.Function(input_names=["T1_file", "NMT_SS_file"],
                         output_names=["aff_file", "warp_file",
                                       "warpinv_file", "transfo_file",
                                       "inv_transfo_file"],
                         function=animal_warper),
            name='NMT_subject_align')
    else:
        print("running default NMT_subject_align version (NMTSubjectAlign2)")

        NMT_subject_align = NodeParams(
            NMTSubjectAlign2(),
            params=parse_key(params, "NMT_subject_align"),
            name='NMT_subject_align')

    NMT_subject_align.inputs.NMT_SS_file = params_template["template_brain"]

    register_NMT_pipe.connect(deoblique, 'out_file',
                              NMT_subject_align, "T1_file")

    # defining list priors
    list_priors = [params_template["template_head"]]

    if "template_seg" in params_template.keys():
        # align_masks
        list_priors.append(params_template["template_seg"])

    else:
        if "template_csf" in params_template.keys():
            list_priors.append(params_template["template_csf"])

        if "template_gm" in params_template.keys():
            list_priors.append(params_template["template_gm"])

        if "template_wm" in params_template.keys():
            list_priors.append(params_template["template_wm"])

    # align_masks
    # "overwrap" of NwarpApply, with specifying the outputs as wished
    align_masks = pe.Node(NwarpApplyPriors(), name='align_masks')
    align_masks.inputs.in_file = list_priors
    align_masks.inputs.out_file = list_priors
    align_masks.inputs.interp = "NN"
    align_masks.inputs.args = "-overwrite"

    register_NMT_pipe.connect(NMT_subject_align, 'aff_file',
                              align_masks, 'master')
    register_NMT_pipe.connect(NMT_subject_align, 'warpinv_file',
                              align_masks, "warp")

    # align_NMT
    align_NMT = pe.Node(
        afni.Allineate(), name="align_NMT", iterfield=['in_file'])
    align_NMT.inputs.final_interpolation = "nearestneighbour"
    align_NMT.inputs.overwrite = True
    align_NMT.inputs.outputtype = "NIFTI_GZ"

    register_NMT_pipe.connect(align_masks, ('out_file', get_elem, 0),
                              align_NMT, "in_file")  # -source

    register_NMT_pipe.connect(deoblique, 'out_file',
                              align_NMT, "reference")  # -base

    register_NMT_pipe.connect(NMT_subject_align, 'inv_transfo_file',
                              align_NMT, "in_matrix")  # -1Dmatrix_apply

    if "template_seg" in params_template.keys():

        # seg
        align_seg = pe.Node(
            afni.Allineate(), name="align_seg", iterfield=['in_file'])
        align_seg.inputs.final_interpolation = "nearestneighbour"
        align_seg.inputs.overwrite = True
        align_seg.inputs.outputtype = "NIFTI_GZ"

        register_NMT_pipe.connect(align_masks, ('out_file', get_elem, 1),
                                  align_seg, "in_file")  # -source
        register_NMT_pipe.connect(deoblique, 'out_file',
                                  align_seg, "reference")  # -base
        register_NMT_pipe.connect(NMT_subject_align, 'inv_transfo_file',
                                  align_seg, "in_matrix")  # -1Dmatrix_apply

    else:
        if "template_csf" in params_template.keys():
            # seg_csf
            align_seg_csf = pe.Node(
                afni.Allineate(), name="align_seg_csf", iterfield=['in_file'])
            align_seg_csf.inputs.final_interpolation = "nearestneighbour"
            align_seg_csf.inputs.overwrite = True
            align_seg_csf.inputs.outputtype = "NIFTI_GZ"

            register_NMT_pipe.connect(align_masks, ('out_file', get_elem, 1),
                                      align_seg_csf, "in_file")  # -source

            register_NMT_pipe.connect(deoblique, 'out_file',
                                      align_seg_csf, "reference")  # -base

            register_NMT_pipe.connect(
                NMT_subject_align, 'inv_transfo_file', align_seg_csf,
                "in_matrix")  # -1Dmatrix_apply

        if "template_gm" in params_template.keys():
            # seg_gm
            align_seg_gm = pe.Node(
                afni.Allineate(), name="align_seg_gm", iterfield=['in_file'])
            align_seg_gm.inputs.final_interpolation = "nearestneighbour"
            align_seg_gm.inputs.overwrite = True
            align_seg_gm.inputs.outputtype = "NIFTI_GZ"

            register_NMT_pipe.connect(align_masks, ('out_file', get_elem, 2),
                                      align_seg_gm, "in_file")  # -source
            register_NMT_pipe.connect(deoblique, 'out_file',
                                      align_seg_gm, "reference")  # -base
            register_NMT_pipe.connect(NMT_subject_align, 'inv_transfo_file',
                                      align_seg_gm, "in_matrix")
        if "template_wm" in params_template.keys():

            # seg_wm
            align_seg_wm = pe.Node(afni.Allineate(), name="align_seg_wm",
                                   iterfield=['in_file'])
            align_seg_wm.inputs.final_interpolation = "nearestneighbour"
            align_seg_wm.inputs.overwrite = True
            align_seg_wm.inputs.outputtype = "NIFTI_GZ"

            register_NMT_pipe.connect(align_masks, ('out_file', get_elem, 3),
                                      align_seg_wm, "in_file")  # -source
            register_NMT_pipe.connect(deoblique, 'out_file',
                                      align_seg_wm, "reference")  # -base
            register_NMT_pipe.connect(NMT_subject_align, 'inv_transfo_file',
                                      align_seg_wm, "in_matrix")

    return register_NMT_pipe


def create_reg_seg_pipe(name="reg_seg_pipe"):

    reg_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['native_seg',
                                      'native_threshold_gm',
                                      'native_threshold_wm',
                                      'native_threshold_csf',
                                      'native_prob_gm',
                                      'native_prob_wm',
                                      'native_prob_csf',
                                      'transfo_file',
                                      'ref_image']),
        name='inputnode')

    # align_seg
    align_seg = pe.Node(
        afni.Allineate(), name="align_seg", iterfield=['in_file'])
    align_seg.inputs.final_interpolation = "nearestneighbour"
    align_seg.inputs.overwrite = True
    align_seg.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_seg',
                     align_seg, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_seg, "reference")
    reg_pipe.connect(inputnode, 'transfo_file', align_seg, "in_matrix")

    # align_threshold_gm
    align_threshold_gm = pe.Node(
        afni.Allineate(), name="align_threshold_gm", iterfield=['in_file'])
    align_threshold_gm.inputs.final_interpolation = "nearestneighbour"
    align_threshold_gm.inputs.overwrite = True
    align_threshold_gm.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_threshold_gm',
                     align_threshold_gm, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_threshold_gm, "reference")
    reg_pipe.connect(inputnode, 'transfo_file',
                     align_threshold_gm, "in_matrix")

    # align_threshold_wm
    align_threshold_wm = pe.Node(
        afni.Allineate(), name="align_threshold_wm", iterfield=['in_file'])
    align_threshold_wm.inputs.final_interpolation = "nearestneighbour"
    align_threshold_wm.inputs.overwrite = True
    align_threshold_wm.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_threshold_wm',
                     align_threshold_wm, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_threshold_wm, "reference")
    reg_pipe.connect(inputnode, 'transfo_file',
                     align_threshold_wm, "in_matrix")

    # align_threshold_csf
    align_threshold_csf = pe.Node(
        afni.Allineate(), name="align_threshold_csf", iterfield=['in_file'])
    align_threshold_csf.inputs.final_interpolation = "nearestneighbour"
    align_threshold_csf.inputs.overwrite = True
    align_threshold_csf.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_threshold_csf',
                     align_threshold_csf, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_threshold_csf, "reference")
    reg_pipe.connect(inputnode, 'transfo_file',
                     align_threshold_csf, "in_matrix")

    # align_prob_gm
    align_prob_gm = pe.Node(
        afni.Allineate(), name="align_prob_gm", iterfield=['in_file'])
    align_prob_gm.inputs.final_interpolation = "nearestneighbour"
    align_prob_gm.inputs.overwrite = True
    align_prob_gm.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_prob_gm',
                     align_prob_gm, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_prob_gm, "reference")
    reg_pipe.connect(inputnode, 'transfo_file', align_prob_gm, "in_matrix")

    # align_prob_wm
    align_prob_wm = pe.Node(
        afni.Allineate(), name="align_prob_wm", iterfield=['in_file'])
    align_prob_wm.inputs.final_interpolation = "nearestneighbour"
    align_prob_wm.inputs.overwrite = True
    align_prob_wm.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_prob_wm',
                     align_prob_wm, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_prob_wm, "reference")
    reg_pipe.connect(inputnode, 'transfo_file', align_prob_wm, "in_matrix")

    # align_prob_csf
    align_prob_csf = pe.Node(
        afni.Allineate(), name="align_prob_csf", iterfield=['in_file'])
    align_prob_csf.inputs.final_interpolation = "nearestneighbour"
    align_prob_csf.inputs.overwrite = True
    align_prob_csf.inputs.outputtype = "NIFTI_GZ"

    reg_pipe.connect(inputnode, 'native_prob_csf',
                     align_prob_csf, "in_file")
    reg_pipe.connect(inputnode, 'ref_image', align_prob_csf, "reference")
    reg_pipe.connect(inputnode, 'transfo_file', align_prob_csf, "in_matrix")

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['norm_seg', 'norm_threshold_gm',
                                      'norm_threshold_wm',
                                      'norm_threshold_csf',
                                      'norm_prob_gm', 'norm_prob_wm',
                                      'norm_prob_csf']),
        name='outputnode')

    reg_pipe.connect(align_seg, 'out_file', outputnode, "norm_seg")

    reg_pipe.connect(align_threshold_gm, 'out_file',
                     outputnode, "norm_threshold_gm")
    reg_pipe.connect(align_threshold_wm, 'out_file',
                     outputnode, "norm_threshold_wm")
    reg_pipe.connect(align_threshold_csf, 'out_file',
                     outputnode, "norm_threshold_csf")

    reg_pipe.connect(align_prob_gm, 'out_file', outputnode, "norm_prob_gm")
    reg_pipe.connect(align_prob_wm, 'out_file', outputnode, "norm_prob_wm")
    reg_pipe.connect(align_prob_csf, 'out_file', outputnode, "norm_prob_csf")

    return reg_pipe



"""
copied from https://stackoverrun.com/fr/q/6768689
seems was never wrapped in nipype
"""
import os
import json

from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec
from nipype.interfaces.base import (CommandLine, CommandLineInputSpec,
                                    TraitedSpec, File, traits)


# FslOrient
class FslOrientInputSpec(FSLCommandInputSpec):

    main_option = traits.Str(
        desc='main option', argstr='-%s', position=0, mandatory=True)

    code = traits.Int(
        argstr='%d', desc='code for setsformcode', position=1)

    in_file = File(
        exists=True, desc='input file', argstr='%s', position=2,
        mandatory=True)


class FslOrientOutputSpec(TraitedSpec):

    out_file = File(desc="out file", exists=True)


class FslOrient(FSLCommand):
    """
    copied and adapted from https://stackoverrun.com/fr/q/6768689
    seems was never wrapped in nipype
    """
    _cmd = 'fslorient'
    input_spec = FslOrientInputSpec
    output_spec = FslOrientOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath(self.inputs.in_file)
        return outputs


# CropVolume
class CropVolumeInputSpec(CommandLineInputSpec):

    # mandatory
    i_file = File(
        exists=True,
        desc='Volume to crop (you can specify as many -in as you want)',
        mandatory=True, position=0, argstr="-i %s")

    b_file = File(
        exists=True,
        desc='Brain image or brain mask, in the same space as the in-file(s)',
        mandatory=True, position=1, argstr="-b %s")

    # optional
    o = traits.String(
        desc="Prefix for the cropped image(s) (Must provide as many prefixes\
            as input images with -o, default is the base name of each input \
            image).",
        argstr="-o %s", mandatory=False)

    s = traits.String(
        "_cropped", usedefault=True,
        desc="Suffix for the cropped image(s) (default is \"_cropped\")",
        argstr="-s %s", mandatory=False)

    c = traits.Int(
        10, usedefault=True,
        desc='c is the space between the brain and the limits of\
            the crop box expressed in percentage of the brain size (eg. if the\
            brain size is 200 voxels in one dimension and c=10: the sides of\
            the brain in this dimension will be 20 voxels away from the\
            borders of the resulting crop box in this dimension). \
            Default: c=10',
        argstr="-c %d", mandatory=False)

    p = traits.String(
        desc="Prefix for running FSL functions\
            (can be a path or just a prefix)",
        argstr="-p %s", mandatory=False)


class CropVolumeOutputSpec(TraitedSpec):
    cropped_file = File(
        exists=True,
        desc="cropped image from CropVolume.sh")


class CropVolume(CommandLine):
    """
    Description: Crop image(s) based on a brain extraction. Multiple images
    can be cropped at once. Will crop each volume preceeded by the -i option

    Inputs:

        Mandatory:

            i_file
                Volume to crop (you can specify as many -in as you want)
            b_file
                Brain image or brain mask, in the same space as the in-file(s)

        Optional:

            o
                Prefix for the cropped image(s) (Must provide as many prefixes
                    as input images with -o, default is the base name of each
                    input image).
            s
                Suffix for the cropped image(s) (default is "_cropped")
            c
                c is the space between the brain and the limits of\
                    the crop box expressed in percentage of the brain size (eg.
                    if the brain size is 200 voxels in one dimension and c=10:
                    the sides of the brain in this dimension will be 20 voxels
                    away from the  borders of the resulting crop box in this
                    dimension). Default: c=10
            p
                Prefix for running FSL functions (can be a path or just a
                prefix)

    Outputs:

        cropped_file:

            cropped image from CropVolume.sh
    """
    input_spec = CropVolumeInputSpec
    output_spec = CropVolumeOutputSpec

    path_to_current_dir = os.path.dirname(os.path.abspath(__file__))
    package_directory = os.path.dirname(path_to_current_dir)

    _cmd = 'bash {}/bash/CropVolume.sh'.format(package_directory)

    def _list_outputs(self):

        import os
        from nipype.utils.filemanip import split_filename as split_f

        outputs = self._outputs().get()

        path, fname, ext = split_f(self.inputs.i_file)

        outfile = fname + self.inputs.s

        if self.inputs.o:
            outfile = self.inputs.o + outfile

        outputs["cropped_file"] = os.path.abspath(outfile + ".nii.gz")
        return outputs


###############################################################################
# Equivalent of flirt_average in FSL
def average_align(list_img):

    import os
    import nibabel as nib
    import numpy as np

    from nipype.utils.filemanip import split_filename as split_f
    import nipype.interfaces.fsl as fsl

    print("average_align:", list_img)

    if isinstance(list_img, list):

        assert len(list_img) > 0, "Error, list should have at least one file"

        if len(list_img) == 1:
            assert os.path.exists(list_img[0])
            av_img_file = list_img[0]
        else:

            img_0 = nib.load(list_img[0])
            path, fname, ext = split_f(list_img[0])

            list_data = [img_0.get_data()]
            for i, img in enumerate(list_img[1:]):

                print("running flirt on {}".format(img))
                flirt = fsl.FLIRT(dof=6)
                flirt.inputs.in_file = img
                flirt.inputs.reference = list_img[0]
                flirt.inputs.interp = "sinc"
                flirt.inputs.no_search = True
                out_file = flirt.run().outputs.out_file
                print(out_file)

                data = nib.load(out_file).get_data()
                list_data.append(data)

            avg_data = np.mean(np.array(list_data), axis=0)
            print(avg_data.shape)

            av_img_file = os.path.abspath("avg_" + fname + ext)
            nib.save(nib.Nifti1Image(avg_data, header=img_0.header,
                                     affine=img_0.affine),
                     av_img_file)

    else:
        assert os.path.exists(list_img)
        av_img_file = list_img

    return av_img_file


# on the fly
def read_cropbox(cropbox_file):

    import os.path as op

    assert op.exists(cropbox_file), \
        "Error {} do not exists".format(cropbox_file)

    with open(cropbox_file) as f:
        crop_list = []
        for line in f.readlines():
            print(line.strip().split())
            crop_list.append(tuple(map(int, map(float, line.strip().split()))))

    return crop_list


def padding_cropped_img(cropped_img_file, orig_img_file, indiv_crop):

    import os

    import nibabel as nib
    import numpy as np

    from nipype.utils.filemanip import split_filename as split_f

    # cropping params
    crop = indiv_crop['crop_T1']['args'].split()
    print("cropping params {}".format(crop))

    xmin = int(crop[0])
    xmax = xmin + int(crop[1])

    ymin = int(crop[2])
    ymax = ymin + int(crop[3])

    zmin = int(crop[4])
    zmax = zmin + int(crop[5])

    # orig image
    orig_img = nib.load(orig_img_file)

    data_orig = orig_img.get_data()
    header_orig = orig_img.header
    affine_orig = orig_img.affine

    print("Orig img shape:", data_orig.shape)

    # cropped image
    cropped_img = nib.load(cropped_img_file)
    data_cropped = cropped_img.get_data()

    print("Cropped img shape:", data_cropped.shape)

    if len(data_orig.shape) == 3 and len(data_cropped.shape) == 4:
        print("Padding with 3D params on a 4D image")

        padded_img_data_shape = (*data_orig.shape, data_cropped.shape[3])
        print(padded_img_data_shape)

        padded_img_data = np.zeros(shape=padded_img_data_shape,
                                   dtype=data_cropped.dtype)
        print("Broscasted padded img shape:", padded_img_data_shape)

        for t in range(data_cropped.shape[3]):
            padded_img_data[xmin:xmax, ymin:ymax, zmin:zmax, t] = \
                data_cropped[:, :, :, t]

    else:

        padded_img_data = np.zeros(shape=data_orig.shape,
                                   dtype=data_cropped.dtype)
        print("Padded img shape:", padded_img_data.shape)

        padded_img_data[xmin:xmax, ymin:ymax, zmin:zmax] = data_cropped

    # saving padded image
    fpath, fname, ext = split_f(cropped_img_file)
    padded_img_file = os.path.abspath(fname + "_padded" + ext)
    img_padded_res = nib.Nifti1Image(padded_img_data, affine=affine_orig,
                                     header=header_orig)
    nib.save(img_padded_res, padded_img_file)

    return padded_img_file


def reg_aladin_dirty(reference, in_file):

    import os

    cwd = os.path.abspath("")
    os.chdir(cwd)
    out_file = os.path.abspath("outputResult.nii")
    cmd = "reg_aladin -flo {} -ref {} -res {}".format(
        in_file, reference, out_file)
    os.system(cmd)

    assert os.path.exists(out_file)
    return out_file


if __name__ == '__main__':

    data_path = "/hpc/meca/data/Macaques/Macaque_hiphop/results/ucdavis"

    orig_file = os.path.join(data_path, "sub-032139/ses-001/anat/\
                             sub-032139_ses-001_run-1_T1w.nii.gz")

    mask_file = os.path.join(
        data_path,
        "derivatives/macapype_ANTS/sub-032139/ses-001/anat/\
        sub-032139_ses-001_space-orig_desc-brain_mask.nii.gz")

    indiv_file = os.path.join(
        "/hpc/meca/users/essamlali.a/Packages",
        "indiv_params_segment_macaque_ants_based_crop.json")
    indiv = json.load(open(indiv_file))

    indiv_crop = indiv['sub-032139']['ses-001']

    res = padding_cropped_img(mask_file, orig_file, indiv_crop)

    print(res)




import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

import nipype.interfaces.fsl as fsl
import nipype.interfaces.afni as afni
import nipype.interfaces.freesurfer as fs

import macapype.nodes.register as reg
from macapype.nodes.surface import Meshify, split_LR_mask
from macapype.utils.utils_nodes import parse_key, NodeParams


def _create_split_hemi_pipe(params, params_template, name="split_hemi_pipe"):

    """Description: Split segmentated tissus according hemisheres after \
    removal of cortical structure

    Processing steps:

    - TODO

    Params:

        - None so far

    Inputs:

        inputnode:

            warpinv_file:
                non-linear transformation (from NMT_subject_align)

            inv_transfo_file:
                inverse transformation

            aff_file:
                affine transformation file

            t1_ref_file:
                preprocessd T1

            segmented_file:
                from atropos segmentation, with all the tissues segmented

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "split_hemi_pipe")

    Outputs:
    """
    split_hemi_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['warpinv_file',
                                      'inv_transfo_file',
                                      'aff_file',
                                      't1_ref_file',
                                      'segmented_file']),
        name='inputnode')

    # get values

    if "cereb_template" in params_template.keys():
        cereb_template_file = params_template["cereb_template"]

        # ### cereb
        # Binarize cerebellum
        bin_cereb = pe.Node(interface=fsl.UnaryMaths(), name='bin_cereb')
        bin_cereb.inputs.operation = "bin"

        bin_cereb.inputs.in_file = cereb_template_file

        # Warp cereb brainmask to subject space
        warp_cereb = pe.Node(interface=reg.NwarpApplyPriors(),
                             name='warp_cereb')

        warp_cereb.inputs.in_file = cereb_template_file
        warp_cereb.inputs.out_file = cereb_template_file
        warp_cereb.inputs.interp = "NN"
        warp_cereb.inputs.args = "-overwrite"

        split_hemi_pipe.connect(bin_cereb, 'out_file', warp_cereb, 'in_file')
        split_hemi_pipe.connect(inputnode, 'aff_file',
                                warp_cereb, 'master')
        split_hemi_pipe.connect(inputnode, 'warpinv_file', warp_cereb, "warp")

        # Align cereb template
        align_cereb = pe.Node(interface=afni.Allineate(), name='align_cereb')

        align_cereb.inputs.final_interpolation = "nearestneighbour"
        align_cereb.inputs.overwrite = True
        align_cereb.inputs.outputtype = "NIFTI_GZ"

        split_hemi_pipe.connect(warp_cereb, 'out_file',
                                align_cereb, "in_file")  # -source
        split_hemi_pipe.connect(inputnode, 't1_ref_file',
                                align_cereb, "reference")  # -base
        split_hemi_pipe.connect(inputnode, 'inv_transfo_file',
                                align_cereb, "in_matrix")  # -1Dmatrix_apply

    if "L_hemi_template" in params_template.keys() and \
            "R_hemi_template" in params_template.keys():

        L_hemi_template_file = params_template["L_hemi_template"]
        R_hemi_template_file = params_template["R_hemi_template"]

        # Warp L hemi template brainmask to subject space
        warp_L_hemi = pe.Node(interface=reg.NwarpApplyPriors(),
                              name='warp_L_hemi')

        warp_L_hemi.inputs.in_file = L_hemi_template_file
        warp_L_hemi.inputs.out_file = L_hemi_template_file
        warp_L_hemi.inputs.interp = "NN"
        warp_L_hemi.inputs.args = "-overwrite"

        split_hemi_pipe.connect(inputnode, 'aff_file',
                                warp_L_hemi, 'master')
        split_hemi_pipe.connect(inputnode, 'warpinv_file', warp_L_hemi, "warp")

        # Align L hemi template
        align_L_hemi = pe.Node(interface=afni.Allineate(), name='align_L_hemi')

        align_L_hemi.inputs.final_interpolation = "nearestneighbour"
        align_L_hemi.inputs.overwrite = True
        align_L_hemi.inputs.outputtype = "NIFTI_GZ"

        split_hemi_pipe.connect(warp_L_hemi, 'out_file',
                                align_L_hemi, "in_file")  # -source
        split_hemi_pipe.connect(inputnode, 't1_ref_file',
                                align_L_hemi, "reference")  # -base
        split_hemi_pipe.connect(inputnode, 'inv_transfo_file',
                                align_L_hemi, "in_matrix")  # -1Dmatrix_apply

        # Warp R hemi template brainmask to subject space
        warp_R_hemi = pe.Node(interface=reg.NwarpApplyPriors(),
                              name='warp_R_hemi')

        warp_R_hemi.inputs.in_file = R_hemi_template_file
        warp_R_hemi.inputs.out_file = R_hemi_template_file
        warp_R_hemi.inputs.interp = "NN"
        warp_R_hemi.inputs.args = "-overwrite"

        split_hemi_pipe.connect(inputnode, 'aff_file',
                                warp_R_hemi, 'master')
        split_hemi_pipe.connect(inputnode, 'warpinv_file', warp_R_hemi, "warp")

        # Align R hemi template
        align_R_hemi = pe.Node(interface=afni.Allineate(), name='align_R_hemi')

        align_R_hemi.inputs.final_interpolation = "nearestneighbour"
        align_R_hemi.inputs.overwrite = True
        align_R_hemi.inputs.outputtype = "NIFTI_GZ"

        split_hemi_pipe.connect(warp_R_hemi, 'out_file',
                                align_R_hemi, "in_file")  # -source
        split_hemi_pipe.connect(inputnode, 't1_ref_file',
                                align_R_hemi, "reference")  # -base
        split_hemi_pipe.connect(inputnode, 'inv_transfo_file',
                                align_R_hemi, "in_matrix")  # -1Dmatrix_apply

    elif "LR_hemi_template" in params_template.keys():

        LR_hemi_template_file = params_template["LR_hemi_template"]

        # Warp LR hemi template brainmask to subject space
        warp_LR_hemi = pe.Node(interface=reg.NwarpApplyPriors(),
                               name='warp_LR_hemi')

        warp_LR_hemi.inputs.in_file = LR_hemi_template_file
        warp_LR_hemi.inputs.out_file = LR_hemi_template_file
        warp_LR_hemi.inputs.interp = "NN"
        warp_LR_hemi.inputs.args = "-overwrite"

        split_hemi_pipe.connect(inputnode, 'aff_file',
                                warp_LR_hemi, 'master')
        split_hemi_pipe.connect(inputnode, 'warpinv_file',
                                warp_LR_hemi, "warp")

        # Align LR hemi template
        align_LR_hemi = pe.Node(interface=afni.Allineate(),
                                name='align_LR_hemi')

        align_LR_hemi.inputs.final_interpolation = "nearestneighbour"
        align_LR_hemi.inputs.overwrite = True
        align_LR_hemi.inputs.outputtype = "NIFTI_GZ"

        split_hemi_pipe.connect(warp_LR_hemi, 'out_file',
                                align_LR_hemi, "in_file")  # -source
        split_hemi_pipe.connect(inputnode, 't1_ref_file',
                                align_LR_hemi, "reference")  # -base
        split_hemi_pipe.connect(inputnode, 'inv_transfo_file',
                                align_LR_hemi, "in_matrix")  # -1Dmatrix_apply

        split_LR = pe.Node(
            interface=niu.Function(
                input_names=["LR_mask_file"],
                output_names=["L_mask_file", "R_mask_file"],
                function=split_LR_mask),
            name="split_LR")

        split_hemi_pipe.connect(align_LR_hemi, "out_file",
                                split_LR, 'LR_mask_file')

    else:
        print("Error, could not find LR_hemi_template or L_hemi_template and \
            R_hemi_template, skipping")
        print(params_template.keys())

        exit()

    # Using LH and RH masks to obtain hemisphere segmentation masks
    calc_L_hemi = pe.Node(interface=afni.Calc(), name='calc_L_hemi')
    calc_L_hemi.inputs.expr = 'a*b/b'
    calc_L_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 'segmented_file',
                            calc_L_hemi, "in_file_a")

    if "LR_hemi_template" in params_template.keys():
        split_hemi_pipe.connect(split_LR, 'L_mask_file',
                                calc_L_hemi, "in_file_b")
    else:
        split_hemi_pipe.connect(align_L_hemi, 'out_file',
                                calc_L_hemi, "in_file_b")

    # R_hemi
    calc_R_hemi = pe.Node(interface=afni.Calc(),
                          name='calc_R_hemi')
    calc_R_hemi.inputs.expr = 'a*b/b'
    calc_R_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 'segmented_file',
                            calc_R_hemi, "in_file_a")

    if "LR_hemi_template" in params_template.keys():

        split_hemi_pipe.connect(split_LR, 'R_mask_file',
                                calc_R_hemi, "in_file_b")
    else:
        split_hemi_pipe.connect(align_R_hemi, 'out_file',
                                calc_R_hemi, "in_file_b")

    # remove cerebellum from left and right brain segmentations
    calc_nocb_L_hemi = pe.Node(interface=afni.Calc(),
                               name='calc_nocb_L_hemi')
    calc_nocb_L_hemi.inputs.expr = '(a*(not (b)))'
    calc_nocb_L_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_L_hemi, 'out_file',
                            calc_nocb_L_hemi, "in_file_a")
    split_hemi_pipe.connect(align_cereb, 'out_file',
                            calc_nocb_L_hemi, "in_file_b")

    calc_nocb_R_hemi = pe.Node(interface=afni.Calc(),
                               name='calc_nocb_R_hemi')
    calc_nocb_R_hemi.inputs.expr = '(a*(not (b)))'
    calc_nocb_R_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_R_hemi, 'out_file',
                            calc_nocb_R_hemi, "in_file_a")
    split_hemi_pipe.connect(align_cereb, 'out_file',
                            calc_nocb_R_hemi, "in_file_b")

    # create L/R GM and WM no-cerebellum masks from subject brain segmentation
    calc_GM_nocb_L_hemi = pe.Node(interface=afni.Calc(),
                                  name='calc_GM_nocb_L_hemi')
    calc_GM_nocb_L_hemi.inputs.expr = 'iszero(a-2)'
    calc_GM_nocb_L_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_nocb_L_hemi, 'out_file',
                            calc_GM_nocb_L_hemi, "in_file_a")

    calc_WM_nocb_L_hemi = pe.Node(interface=afni.Calc(),
                                  name='calc_WM_nocb_L_hemi')
    calc_WM_nocb_L_hemi.inputs.expr = 'iszero(a-3)'
    calc_WM_nocb_L_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_nocb_L_hemi, 'out_file',
                            calc_WM_nocb_L_hemi, "in_file_a")

    calc_GM_nocb_R_hemi = pe.Node(interface=afni.Calc(),
                                  name='calc_GM_nocb_R_hemi')
    calc_GM_nocb_R_hemi.inputs.expr = 'iszero(a-2)'
    calc_GM_nocb_R_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_nocb_R_hemi, 'out_file',
                            calc_GM_nocb_R_hemi, "in_file_a")

    calc_WM_nocb_R_hemi = pe.Node(interface=afni.Calc(),
                                  name='calc_WM_nocb_R_hemi')
    calc_WM_nocb_R_hemi.inputs.expr = 'iszero(a-3)'
    calc_WM_nocb_R_hemi.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(calc_nocb_R_hemi, 'out_file',
                            calc_WM_nocb_R_hemi, "in_file_a")

    # Extract Cerebellum using template mask transformed to subject space
    extract_cereb = pe.Node(interface=afni.Calc(), name='extract_cereb')
    extract_cereb.inputs.expr = 'a*b/b'
    extract_cereb.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 't1_ref_file',
                            extract_cereb, "in_file_a")
    split_hemi_pipe.connect(align_cereb, 'out_file',
                            extract_cereb, "in_file_b")

    # Extract L.GM using template mask transformed to subject space
    extract_L_GM = pe.Node(interface=afni.Calc(), name='extract_L_GM')
    extract_L_GM.inputs.expr = 'a*b/b'
    extract_L_GM.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 't1_ref_file',
                            extract_L_GM, "in_file_a")
    split_hemi_pipe.connect(calc_GM_nocb_L_hemi, 'out_file',
                            extract_L_GM, "in_file_b")

    # Extract L.WM using template mask transformed to subject space
    extract_L_WM = pe.Node(interface=afni.Calc(), name='extract_L_WM')
    extract_L_WM.inputs.expr = 'a*b/b'
    extract_L_WM.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 't1_ref_file',
                            extract_L_WM, "in_file_a")
    split_hemi_pipe.connect(calc_WM_nocb_L_hemi, 'out_file',
                            extract_L_WM, "in_file_b")

    # Extract L.GM using template mask transformed to subject space
    extract_R_GM = pe.Node(interface=afni.Calc(), name='extract_R_GM')
    extract_R_GM.inputs.expr = 'a*b/b'
    extract_R_GM.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 't1_ref_file',
                            extract_R_GM, "in_file_a")
    split_hemi_pipe.connect(calc_GM_nocb_R_hemi, 'out_file',
                            extract_R_GM, "in_file_b")

    # Extract L.WM using template mask transformed to subject space
    extract_R_WM = pe.Node(interface=afni.Calc(), name='extract_R_WM')
    extract_R_WM.inputs.expr = 'a*b/b'
    extract_R_WM.inputs.outputtype = 'NIFTI_GZ'

    split_hemi_pipe.connect(inputnode, 't1_ref_file',
                            extract_R_WM, "in_file_a")
    split_hemi_pipe.connect(calc_WM_nocb_R_hemi, 'out_file',
                            extract_R_WM, "in_file_b")

    return split_hemi_pipe


def create_nii_to_mesh_pipe(params, params_template, name="nii_to_mesh_pipe"):
    """
    Description: basic nii to mesh pipeline after segmentation

    Processing steps:

    - split in hemisphere after removal of the subcortical structures
    - using scipy marching cube methods to compute mesh
    - some smoothing (using brain-slam)

    Params:

        - split_hemi_pipe (see :class:`_create_split_hemi_pipe \
        <macapype.pipelines.surface._create_split_hemi_pipe>` for arguments)
        - mesh_L_GM, mesh_R_GM, mesh_L_WM, mesh_R_WM (see :class:`Meshify  \
        <macapype.nodes.surface.Meshify>` for arguments)

    Inputs:

        inputnode:

            warpinv_file:
                non-linear transformation (from NMT_subject_align)

            inv_transfo_file:
                inverse transformation

            aff_file:
                affine transformation file

            t1_ref_file:
                preprocessd T1

            segmented_file:
                from atropos segmentation, with all the tissues segmented

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "nii_to_mesh_pipe")

    Outputs:
    """
    # creating pipeline
    mesh_to_seg_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['warpinv_file',
                                      'inv_transfo_file',
                                      'aff_file',
                                      't1_ref_file',
                                      'segmented_file']),
        name='inputnode')

    # split hemi pipe (+ cerebellum)
    split_hemi_pipe = _create_split_hemi_pipe(
        params=parse_key(params, "split_hemi_pipe"),
        params_template=params_template)

    mesh_to_seg_pipe.connect(inputnode, 'warpinv_file',
                             split_hemi_pipe, 'inputnode.warpinv_file')

    mesh_to_seg_pipe.connect(inputnode, 'inv_transfo_file',
                             split_hemi_pipe, 'inputnode.inv_transfo_file')

    mesh_to_seg_pipe.connect(inputnode, 'aff_file',
                             split_hemi_pipe, 'inputnode.aff_file')

    mesh_to_seg_pipe.connect(inputnode, 't1_ref_file',
                             split_hemi_pipe, 'inputnode.t1_ref_file')

    mesh_to_seg_pipe.connect(inputnode, 'segmented_file',
                             split_hemi_pipe, 'inputnode.segmented_file')

    # meshify L GM hemisphere
    mesh_L_GM = NodeParams(interface=Meshify(),
                           params=parse_key(params, "mesh_L_GM"),
                           name="mesh_L_GM")

    mesh_to_seg_pipe.connect(split_hemi_pipe, 'extract_L_GM.out_file',
                             mesh_L_GM, "image_file")

    # meshify R GM hemisphere
    mesh_R_GM = NodeParams(interface=Meshify(),
                           params=parse_key(params, "mesh_R_GM"),
                           name="mesh_R_GM")

    mesh_to_seg_pipe.connect(split_hemi_pipe, 'extract_R_GM.out_file',
                             mesh_R_GM, "image_file")

    # meshify L WM hemisphere
    mesh_L_WM = NodeParams(interface=Meshify(),
                           params=parse_key(params, "mesh_L_WM"),
                           name="mesh_L_WM")

    mesh_to_seg_pipe.connect(split_hemi_pipe, 'extract_L_WM.out_file',
                             mesh_L_WM, "image_file")

    # meshify R WM hemisphere
    mesh_R_WM = NodeParams(interface=Meshify(),
                           params=parse_key(params, "mesh_R_WM"),
                           name="mesh_R_WM")

    mesh_to_seg_pipe.connect(split_hemi_pipe, 'extract_R_WM.out_file',
                             mesh_R_WM, "image_file")

    return mesh_to_seg_pipe


###############################################################################
# using freesurfer tools to build the mesh through tesselation
def create_nii_to_mesh_fs_pipe(params, name="nii_to_mesh_fs_pipe"):

    """
    Description: surface generation using freesurfer tools

    Params:

    - fill_wm (see `MRIFill <https://nipype.readthedocs.io/en/0.12.1/\
    interfaces/generated/nipype.interfaces.freesurfer.utils.html#mrifill>`_) \
    - also available as :ref:`indiv_params <indiv_params>`


    Inputs:

        inputnode:

            wm_mask_file:
                segmented white matter mask (binary) in template space

            reg_brain_file:
                preprocessd T1, registered to template

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "nii_to_mesh_fs_pipe")

    Outputs:

    """
    # creating pipeline
    nii_to_mesh_fs_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['wm_mask_file', 'reg_brain_file',
                                      'indiv_params']),
        name='inputnode')

    # bin_wm
    bin_wm = pe.Node(interface=fsl.UnaryMaths(), name="bin_wm")
    bin_wm.inputs.operation = "fillh"

    nii_to_mesh_fs_pipe.connect(inputnode, 'wm_mask_file',
                                bin_wm, 'in_file')

    # resample everything
    refit_wm = pe.Node(interface=afni.Refit(), name="refit_wm")
    refit_wm.inputs.args = "-xdel 1.0 -ydel 1.0 -zdel 1.0 -keepcen"

    nii_to_mesh_fs_pipe.connect(bin_wm, 'out_file',
                                refit_wm, 'in_file')

    # resample everything
    refit_reg = pe.Node(interface=afni.Refit(), name="refit_reg")
    refit_reg.inputs.args = "-xdel 1.0 -ydel 1.0 -zdel 1.0 -keepcen"

    nii_to_mesh_fs_pipe.connect(inputnode, 'reg_brain_file',
                                refit_reg, 'in_file')

    # mri_convert wm to freesurfer mgz
    convert_wm = pe.Node(interface=fs.MRIConvert(),
                         name="convert_wm")
    convert_wm.inputs.out_type = "mgz"
    convert_wm.inputs.conform = True

    nii_to_mesh_fs_pipe.connect(refit_wm, 'out_file',
                                convert_wm, 'in_file')

    # mri_convert reg to freesurfer mgz
    convert_reg = pe.Node(interface=fs.MRIConvert(),
                          name="convert_reg")
    convert_reg.inputs.out_type = "mgz"
    convert_reg.inputs.conform = True

    nii_to_mesh_fs_pipe.connect(refit_reg, 'out_file',
                                convert_reg, 'in_file')

    # mri_fill
    fill_wm = NodeParams(interface=fs.MRIFill(),
                         params=parse_key(params, "fill_wm"),
                         name="fill_wm")

    fill_wm.inputs.out_file = "filled.mgz"

    nii_to_mesh_fs_pipe.connect(convert_wm, 'out_file',
                                fill_wm, 'in_file')

    nii_to_mesh_fs_pipe.connect(
        inputnode, ("indiv_params", parse_key, "fill_wm"),
        fill_wm, 'indiv_params')

    # pretesselate wm
    pretess_wm = pe.Node(interface=fs.MRIPretess(),
                         name="pretess_wm")
    pretess_wm.inputs.label = 255

    nii_to_mesh_fs_pipe.connect(fill_wm, 'out_file',
                                pretess_wm, 'in_filled')

    nii_to_mesh_fs_pipe.connect(convert_reg, 'out_file',
                                pretess_wm, 'in_norm')

    # tesselate wm lh
    tess_wm_lh = pe.Node(interface=fs.MRITessellate(), name="tess_wm_lh")
    tess_wm_lh.inputs.label_value = 255
    tess_wm_lh.inputs.out_file = "lh_tess"

    nii_to_mesh_fs_pipe.connect(pretess_wm, 'out_file',
                                tess_wm_lh, 'in_file')

    # tesselate wm rh
    tess_wm_rh = pe.Node(interface=fs.MRITessellate(), name="tess_wm_rh")
    tess_wm_rh.inputs.label_value = 127
    tess_wm_rh.inputs.out_file = "rh_tess"

    nii_to_mesh_fs_pipe.connect(pretess_wm, 'out_file',
                                tess_wm_rh, 'in_file')

    # ExtractMainComponent lh
    extract_mc_lh = pe.Node(interface=fs.ExtractMainComponent(),
                            name="extract_mc_lh")

    nii_to_mesh_fs_pipe.connect(tess_wm_lh, 'surface',
                                extract_mc_lh, 'in_file')

    extract_mc_lh.inputs.out_file = "lh.lh_tess.maincmp"

    # ExtractMainComponent rh
    extract_mc_rh = pe.Node(interface=fs.ExtractMainComponent(),
                            name="extract_mc_rh")

    nii_to_mesh_fs_pipe.connect(tess_wm_rh, 'surface',
                                extract_mc_rh, 'in_file')

    extract_mc_rh.inputs.out_file = "rh.rh_tess.maincmp"

    # SmoothTessellation lh
    smooth_tess_lh = pe.Node(interface=fs.SmoothTessellation(),
                             name="smooth_tess_lh")
    smooth_tess_lh.inputs.disable_estimates = True

    nii_to_mesh_fs_pipe.connect(extract_mc_lh, 'out_file',
                                smooth_tess_lh, 'in_file')

    # SmoothTessellation rh
    smooth_tess_rh = pe.Node(interface=fs.SmoothTessellation(),
                             name="smooth_tess_rh")
    smooth_tess_rh.inputs.disable_estimates = True

    nii_to_mesh_fs_pipe.connect(extract_mc_rh, 'out_file',
                                smooth_tess_rh, 'in_file')

    return nii_to_mesh_fs_pipe



from nipype.interfaces.base import TraitedSpec, SimpleInterface
from nipype.interfaces.base import traits, File


def split_LR_mask(LR_mask_file, left_index=1, right_index=2):

    import os

    import nibabel as nib
    import numpy as np

    from nipype.utils.filemanip import split_filename as split_f

    path, fname, ext = split_f(LR_mask_file)

    LR_mask_img = nib.load(LR_mask_file)

    LR_mask_data = LR_mask_img.get_fdata()

    # L_mask
    L_mask_data = np.zeros(shape=LR_mask_data.shape)

    L_mask_data[LR_mask_data == left_index] = 1

    L_mask_file = os.path.abspath("L_mask"+ext)

    nib.save(
        nib.Nifti1Image(L_mask_data, affine=LR_mask_img.affine,
                        header=LR_mask_img.header),
        L_mask_file)

    # R_mask
    R_mask_data = np.zeros(shape=LR_mask_data.shape)

    R_mask_data[LR_mask_data == right_index] = 1

    R_mask_file = os.path.abspath("R_mask"+ext)

    nib.save(
        nib.Nifti1Image(R_mask_data, affine=LR_mask_img.affine,
                        header=LR_mask_img.header),
        R_mask_file)

    return L_mask_file, R_mask_file


class MeshifyInputSpec(TraitedSpec):
    image_file = File(
        exists=True, desc='input file', mandatory=True)

    level = traits.Float(
        0.5, usedefault=True,
        desc='Threshold used to detect surfaces',
        mandatory=False)

    smoothing_iter = traits.Int(
        0, usedefault=True,
        desc='Number of Laplacian smoothing iterations',
        mandatory=False)

    smoothing_dt = traits.Float(
        0.1, usedefault=True,
        desc='dt param of the Laplacian smoothing',
        mandatory=False)


class MeshifyOutputSpec(TraitedSpec):
    mesh_file = File(desc="out file", exists=True)


class Meshify(SimpleInterface):
    """
    Definition: Create a Gifti mesh from a binary Nifti image
    """
    _cmd = 'fslorient'
    input_spec = MeshifyInputSpec
    output_spec = MeshifyOutputSpec

    def _run_interface(self, runtime):
        import os.path as op
        import nibabel.gifti as ng
        import numpy as np
        import skimage.measure as sm
        import nilearn.image as nimg

        import slam.io as sio
        import slam.differential_geometry as sdg

        from nipype.utils.filemanip import split_filename

        # Generate output mesh filename from the input image name
        _, fname, _ = split_filename(self.inputs.image_file)
        gii_file = op.abspath(op.join(runtime.cwd, fname + ".gii"))

        # Load the largest connected component of the input image
        img = nimg.largest_connected_component_img(self.inputs.image_file)

        # TODO: check if the input image is correct (binary)

        # Run the marching cube algorithm
        verts, faces, normals, values = sm.marching_cubes_lewiner(
            img.get_data(), self.inputs.level)

        # Convert vertices coordinates to image space
        # TODO: check that is correct by plotting the mesh on the image
        x, y, z = nimg.coord_transform(
            verts[:, 0], verts[:, 1], verts[:, 2], img.affine)
        mm_verts = np.array([x, y, z]).T

        # Save the mesh as Gifti
        # FIXME: FreeView can not open the mesh (but anatomist do)
        gii = ng.GiftiImage(darrays=[
            ng.GiftiDataArray(mm_verts, intent='NIFTI_INTENT_POINTSET'),
            ng.GiftiDataArray(faces, intent='NIFTI_INTENT_TRIANGLE')])
        gii.meta = ng.GiftiMetaData().from_dict({
            "volume_file": self.inputs.image_file,
            "marching_cube_level": str(self.inputs.level),
            "smoothing_iterations": str(self.inputs.smoothing_iter),
            "smoothing_dt": str(self.inputs.smoothing_dt)
        })
        ng.write(gii, gii_file)

        # Optional: Smooth the marching cube output with SLAM
        if self.inputs.smoothing_iter > 0:
            mesh = sdg.laplacian_mesh_smoothing(
                sio.load_mesh(gii_file),
                nb_iter=self.inputs.smoothing_iter,
                dt=self.inputs.smoothing_dt)
            sio.write_mesh(mesh, gii_file)

        return runtime



"""
    Pipelines for brain extraction

"""
import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

import nipype.interfaces.fsl as fsl
import nipype.interfaces.afni as afni

from ..nodes.extract_brain import AtlasBREX

from ..utils.utils_nodes import NodeParams, parse_key


def create_extract_pipe(params_template, params={}, name="extract_pipe"):
    """
    Description: Extract T1 brain using AtlasBrex

    Params:

    - norm_intensity (see `N4BiasFieldCorrection <https://nipype.readthedocs\
    .io/en/0.12.1/interfaces/generated/nipype.interfaces.ants.segmentation.\
    html#n4biasfieldcorrection>`_ for arguments)
    - atlas_brex (see :class:`AtlasBREX \
    <macapype.nodes.extract_brain.AtlasBREX>` for arguments) - also \
    available as :ref:`indiv_params <indiv_params>`

    Inputs:

        inputnode:

            restore_T1:
                preprocessed (debiased/denoised) T1 file name

            restore_T2:
                preprocessed (debiased/denoised)T2 file name

        arguments:

            params_template:
                dictionary of info about template

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "extract_pipe")

    Outputs:

        smooth_mask.out_file:
            Computed mask (after some smoothing)

    """

    # creating pipeline
    extract_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['restore_T1', 'restore_T2',
                                      "indiv_params"]),
        name='inputnode')

    # atlas_brex
    atlas_brex = NodeParams(AtlasBREX(),
                            params=parse_key(params, "atlas_brex"),
                            name='atlas_brex')

    extract_pipe.connect(inputnode, "restore_T1",
                         atlas_brex, 't1_restored_file')

    atlas_brex.inputs.NMT_file = params_template["template_head"]
    atlas_brex.inputs.NMT_SS_file = params_template["template_brain"]

    extract_pipe.connect(
            inputnode, ("indiv_params", parse_key, "atlas_brex"),
            atlas_brex, 'indiv_params')

    # mask_brex
    mask_brex = pe.Node(fsl.UnaryMaths(), name='mask_brex')
    mask_brex.inputs.operation = 'bin'

    extract_pipe.connect(atlas_brex, 'brain_file', mask_brex, 'in_file')

    # smooth_mask
    smooth_mask = pe.Node(fsl.UnaryMaths(), name='smooth_mask')
    smooth_mask.inputs.operation = "bin"
    smooth_mask.inputs.args = "-s 1 -thr 0.5 -bin"

    extract_pipe.connect(mask_brex, 'out_file', smooth_mask, 'in_file')

    # mult_T1
    mult_T1 = pe.Node(afni.Calc(), name='mult_T1')
    mult_T1.inputs.expr = "a*b"
    mult_T1.inputs.outputtype = 'NIFTI_GZ'

    extract_pipe.connect(inputnode, "restore_T1", mult_T1, 'in_file_a')
    extract_pipe.connect(smooth_mask, 'out_file', mult_T1, 'in_file_b')

    # mult_T2
    mult_T2 = pe.Node(afni.Calc(), name='mult_T2')
    mult_T2.inputs.expr = "a*b"
    mult_T2.inputs.outputtype = 'NIFTI_GZ'

    extract_pipe.connect(inputnode, 'restore_T2', mult_T2, 'in_file_a')
    extract_pipe.connect(smooth_mask, 'out_file', mult_T2, 'in_file_b')
    return extract_pipe


def create_extract_T1_pipe(params_template, params={},
                           name="extract_T1_pipe"):
    """
    Description: Extract T1 brain using AtlasBrex

    Params:

    - norm_intensity (see `N4BiasFieldCorrection <https://nipype.readthedocs.\
    io/en/0.12.1/interfaces/generated/nipype.interfaces.ants.segmentation.html\
    #n4biasfieldcorrection>`_ for arguments)
    - atlas_brex (see :class:`AtlasBREX \
    <macapype.nodes.extract_brain.AtlasBREX>` for arguments) - also available \
    as :ref:`indiv_params <indiv_params>`

    Inputs:

        inputnode:

            restore_T1:
                preprocessed (debiased/denoised) T1 file name

        arguments:

            params_template:
                dictionary of info about template

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "extract_pipe")

    Outputs:

        smooth_mask.out_file:
            Computed mask (after some smoothing)

    """

    # creating pipeline
    extract_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['restore_T1',
                                      "indiv_params"]),
        name='inputnode')

    # atlas_brex
    atlas_brex = NodeParams(AtlasBREX(),
                            params=parse_key(params, "atlas_brex"),
                            name='atlas_brex')

    extract_pipe.connect(inputnode, 'restore_T1',
                         atlas_brex, 't1_restored_file')

    atlas_brex.inputs.NMT_file = params_template["template_head"]
    atlas_brex.inputs.NMT_SS_file = params_template["template_brain"]

    extract_pipe.connect(
            inputnode, ("indiv_params", parse_key, "atlas_brex"),
            atlas_brex, 'indiv_params')

    # mask_brex
    mask_brex = pe.Node(fsl.UnaryMaths(), name='mask_brex')
    mask_brex.inputs.operation = 'bin'

    extract_pipe.connect(atlas_brex, 'brain_file', mask_brex, 'in_file')

    # smooth_mask
    smooth_mask = pe.Node(fsl.UnaryMaths(), name='smooth_mask')
    smooth_mask.inputs.operation = "bin"
    smooth_mask.inputs.args = "-s 1 -thr 0.5 -bin"

    extract_pipe.connect(mask_brex, 'out_file', smooth_mask, 'in_file')

    # mult_T1
    mult_T1 = pe.Node(afni.Calc(), name='mult_T1')
    mult_T1.inputs.expr = "a*b"
    mult_T1.inputs.outputtype = 'NIFTI_GZ'

    extract_pipe.connect(inputnode, 'restore_T1', mult_T1, 'in_file_a')
    extract_pipe.connect(smooth_mask, 'out_file', mult_T1, 'in_file_b')

    return extract_pipe
